@@ -11,15 +11,23 @@
 open Declarations
 
 let set_local_flags flags env =
-  let flags =
-    { (Environ.typing_flags env) with
-      check_guarded = flags.check_guarded;
-      check_positive = flags.check_positive;
-      check_universes = flags.check_universes;
-      conv_oracle = flags.conv_oracle;
-      share_reduction = flags.share_reduction;
-      cumulative_sprop = flags.cumulative_sprop;
-      allow_uip = flags.allow_uip;
-    }
+  (* Explicitly ignored flags are set to not change *)
+  let envflags = Environ.typing_flags env in
+  let flags = {
+    (* These flags may be overridden *)
+    check_guarded = flags.check_guarded;
+    check_positive = flags.check_positive;
+    check_universes = flags.check_universes;
+    conv_oracle = flags.conv_oracle;
+    share_reduction = flags.share_reduction;
+    cumulative_sprop = flags.cumulative_sprop;
+    allow_uip = flags.allow_uip;
+    (* These flags may not *)
+    enable_VM = envflags.enable_VM;
+    enable_native_compiler = envflags.enable_native_compiler;
+    indices_matter = envflags.indices_matter;
+    impredicative_set = envflags.impredicative_set;
+    sprop_allowed = envflags.sprop_allowed;
+  }
   in
   Environ.set_typing_flags flags env
