@@ -71,7 +71,7 @@ val finished : proofview -> bool
 val return : proofview -> Evd.evar_map
 
 val partial_proof : entry -> proofview -> constr list
-val initial_goals : entry -> (constr * types) list
+val initial_goals : entry -> (Environ.named_context_val * constr * types) list
 
 (** goal <-> goal_with_state *)
 
@@ -553,7 +553,6 @@ module Goal : sig
 
   (** Compatibility: avoid if possible *)
   val goal : t -> Evar.t
-  val print : t -> Evar.t Evd.sigma
 
 end
 
@@ -584,44 +583,29 @@ module NonLogical : module type of Logic_monad.NonLogical
 (** [tclLIFT c] is a tactic which behaves exactly as [c]. *)
 val tclLIFT : 'a NonLogical.t -> 'a tactic
 
+(* transforms every Ocaml (catchable) exception into a failure in
+    the monad. *)
+val wrap_exceptions : (unit -> 'a tactic) -> 'a tactic
 
 (**/**)
 
 (*** Compatibility layer with <= 8.2 tactics ***)
 module V82 : sig
+  [@@@ocaml.warning "-3"]
   type tac = Evar.t Evd.sigma -> Evar.t list Evd.sigma
+  [@@@ocaml.warning "+3"]
 
   (* [nf_evars=true] applies the evar (assignment) map to the goals
    * (conclusion and context) before calling the tactic *)
   val tactic : ?nf_evars:bool -> tac -> unit tactic
 
-  (* normalises the evars in the goals, and stores the result in
-     solution. *)
-  val nf_evar_goals : unit tactic
-
-  val has_unresolved_evar : proofview -> bool
-
-  val top_goals : entry -> proofview -> Evar.t list Evd.sigma
-
-  (* returns the existential variable used to start the proof *)
-  val top_evars : entry -> proofview -> Evar.t list
-
   (* Caution: this function loses quite a bit of information. It
      should be avoided as much as possible.  It should work as
      expected for a tactic obtained from {!V82.tactic} though. *)
   val of_tactic : 'a tactic -> tac
 
-  (* marks as unsafe if the argument is [false] *)
-  val put_status : bool -> unit tactic
-
-  (* exception for which it is deemed to be safe to transmute into
-     tactic failure. *)
-  val catchable_exception : exn -> bool
-
-  (* transforms every Ocaml (catchable) exception into a failure in
-     the monad. *)
-  val wrap_exceptions : (unit -> 'a tactic) -> 'a tactic
 end
+[@@ocaml.deprecated "Use the new engine."]
 
 (** {7 Notations} *)
 
