@@ -14,8 +14,20 @@ open Ideutils
 open Interface
 open Feedback
 
-let b2c = byte_off_to_buffer_off
-let c2b = buffer_off_to_byte_off
+let b2c = byte_offset_to_char_offset
+
+(** variant of buffer_off_to_byte_off: convert unicode offset (used by GTK buffer)
+    to UTF-8 byte offset (used by Coq) with size caching for performance *)
+let c2b (buffer : GText.buffer) (s_byte, s_uni) uni_off =
+  if s_uni < 0 then
+    raise (invalid_arg "c2b: s_uni must be >= 0");
+  let rec cvt iter rv =
+    if iter#offset <= s_uni then
+      rv
+    else
+      cvt iter#backward_char (rv + (ulen iter#char))
+  in
+  cvt (buffer#get_iter (`OFFSET uni_off)) s_byte
 
 type flag = [ `INCOMPLETE | `UNSAFE | `PROCESSING | `ERROR of string Loc.located | `WARNING of string Loc.located ]
 type mem_flag = [ `INCOMPLETE | `UNSAFE | `PROCESSING | `ERROR | `WARNING ]
@@ -38,9 +50,10 @@ end
 
 module SentenceId : sig
 
-  type sentence = private {
+  type sentence = {
     start : GText.mark;
     stop : GText.mark;
+    mutable byte_start : int;
     mutable flags : flag list;
     mutable tooltips : (int * int * string) list;
     edit_id : int;
@@ -68,6 +81,7 @@ end = struct
   type sentence = {
     start : GText.mark;
     stop : GText.mark;
+    mutable byte_start : int;
     mutable flags : flag list;
     mutable tooltips : (int * int * string) list;
     edit_id : int;
@@ -85,6 +99,7 @@ end = struct
   let mk_sentence ~start ~stop flags = decr id; {
     start = start;
     stop = stop;
+    byte_start = 0;
     flags = flags;
     edit_id = !id;
     tooltips = [];
@@ -165,7 +180,7 @@ object
   method get_n_errors : int
   method get_errors : (int * string) list
   method get_slaves_status : int * int * string CString.Map.t
-
+  method backtrack_to_begin : unit -> unit task
   method handle_failure : handle_exn_rty -> unit task
 
   method destroy : unit -> unit
@@ -238,6 +253,12 @@ type queue_msg =
   | MsgFeedback of Feedback.feedback
   | MsgDebug of string * Pp.t
 
+let rec flatten = function
+| [] -> []
+| (lg, rg) :: l ->
+  let inner = flatten l in
+  List.rev_append lg inner @ rg
+
 class coqops
   (_script:Wg_ScriptView.script_view)
   (_pv:Wg_ProofView.proof_view)
@@ -262,6 +283,8 @@ object(self)
   val mutable processed = 0
   val mutable slaves_status = CString.Map.empty
 
+  val mutable last_offsets = (0,0)
+
   val mutable forward_clear_db_highlight = ((fun x -> failwith "forward_clear_db_highlight")
                   : unit -> unit)
 
@@ -313,9 +336,7 @@ object(self)
             try Doc.find document mem_marks
             with Not_found -> aux iter#backward_char in
         aux iter in
-      let ss =
-        find_all_tooltips s
-          (iter#offset - (buffer#get_iter_at_mark s.start)#offset) in
+      let ss = find_all_tooltips s iter#offset in
       let msg = String.concat "\n" (CList.uniquize ss) in
       GtkBase.Tooltip.set_icon_from_stock tooltip `INFO `BUTTON;
       script#misc#set_tooltip_markup ("<tt>" ^ msg ^ "</tt>")
@@ -380,18 +401,37 @@ object(self)
         script#recenter_insert
       end
     end;
-    Coq.bind (Coq.goals ()) (function
+    let flags = { gf_mode = "full"; gf_fg = true; gf_bg = true; gf_shelved = false; gf_given_up = false } in
+    let return x = Coq.return (Good x) in
+    let (>>=) m f = Coq.bind m (function
+    | Fail x -> Coq.return (Fail x)
+    | Good v -> f v)
+    in
+    let call =
+      Coq.subgoals flags >>= begin function
+      | None -> return Wg_ProofView.NoGoals
+      | Some { fg_goals = ((_ :: _) as fg); bg_goals = bg } ->
+        let bg = flatten (List.rev bg) in
+        return (Wg_ProofView.FocusGoals { fg; bg; })
+      | Some { fg_goals = []; bg_goals = bg } ->
+        let flags = { gf_mode = "short"; gf_fg = false; gf_bg = false; gf_shelved = true; gf_given_up = true } in
+        Coq.subgoals flags >>= fun rem ->
+        let bg = flatten (List.rev bg) in
+        let shelved, given_up = match rem with
+        | None -> [], []
+        | Some goals -> goals.shelved_goals, goals.given_up_goals
+        in
+        return (Wg_ProofView.NoFocusGoals { bg; shelved; given_up })
+      end
+    in
+    Coq.bind call begin function
     | Fail x -> self#handle_failure_aux ~move_insert x
     | Good goals ->
-      Coq.bind (Coq.evars ()) (function
-        | Fail x -> self#handle_failure_aux ~move_insert x
-        | Good evs ->
-          proof#set_goals goals;
-          proof#set_evars evs;
-          proof#refresh ~force:true;
-          Coq.return ()
-        )
-      )
+      proof#set_goals goals;
+      proof#refresh ~force:true;
+      Coq.return ()
+    end
+
   method show_goals = self#show_goals_aux ()
 
   (* This method is intended to perform stateless commands *)
@@ -454,14 +494,14 @@ object(self)
 
   method private attach_tooltip ?loc sentence text =
     let start_sentence, stop_sentence, phrase = self#get_sentence sentence in
-    let pre_chars, post_chars = Option.cata Loc.unloc (0, String.length phrase) loc in
-    let pre = b2c buffer pre_chars in
-    let post = b2c buffer post_chars in
+    let pre_bytes, post_bytes = Option.cata Loc.unloc (0, String.length phrase) loc in
+    let pre = b2c phrase (pre_bytes - (sentence.byte_start)) in
+    let post = b2c phrase (post_bytes - (sentence.byte_start)) in
     let start = start_sentence#forward_chars pre in
     let stop = start_sentence#forward_chars post in
     let markup = Glib.Markup.escape_text text in
     buffer#apply_tag Tags.Script.tooltip ~start ~stop;
-    add_tooltip sentence pre post markup
+    add_tooltip sentence start#offset stop#offset markup
 
   method private debug_prompt ~tag msg =
     match tag with
@@ -587,27 +627,28 @@ object(self)
     let stop = buffer#get_iter_at_mark sentence.stop in
     buffer#move_mark ~where:stop (`NAME "start_of_input");
 
-  method private position_tag_at_iter ?loc iter_start iter_stop tag phrase = match loc with
+  method private apply_tag_in_sentence ?loc tag sentence =
+    let start, stop, phrase = self#get_sentence sentence in
+    match loc with
     | None ->
-      buffer#apply_tag tag ~start:iter_start ~stop:iter_stop
+      buffer#apply_tag tag ~start ~stop
     | Some loc ->
-      let start, stop = Loc.unloc loc in
-      buffer#apply_tag tag
-        ~start:(buffer#get_iter (`OFFSET (b2c buffer start)))
-        ~stop:(buffer#get_iter (`OFFSET (b2c buffer stop)))
+      let start_bytes, stop_bytes = Loc.unloc loc in
+      let stop = start#forward_chars (b2c phrase (stop_bytes - (sentence.byte_start))) in
+      let start = start#forward_chars (b2c phrase (start_bytes - (sentence.byte_start))) in
+      buffer#apply_tag tag ~start ~stop
 
   method private position_tag_at_sentence ?loc tag sentence =
-    let start, stop, phrase = self#get_sentence sentence in
-    self#position_tag_at_iter ?loc start stop tag phrase
+    self#apply_tag_in_sentence ?loc tag sentence
 
   method private process_interp_error ?loc queue sentence msg tip id =
     Coq.bind (Coq.return ()) (function () ->
-    let start, stop, phrase = self#get_sentence sentence in
+    let start, stop, _ = self#get_sentence sentence in
     buffer#remove_tag Tags.Script.to_process ~start ~stop;
     self#discard_command_queue queue;
     pop_info ();
     if Stateid.equal id tip || Stateid.equal id Stateid.dummy then begin
-      self#position_tag_at_iter ?loc start stop Tags.Script.error phrase;
+      self#apply_tag_in_sentence ?loc Tags.Script.error sentence;
       buffer#place_cursor ~where:stop;
       messages#default_route#clear;
       messages#default_route#push Feedback.Error msg;
@@ -628,7 +669,8 @@ object(self)
   method private fill_command_queue until queue =
     let topstack =
       if Doc.focused document then fst (Doc.context document) else [] in
-    let rec loop n iter =
+    let rec loop n iter prev_start =
+      (* prev_start is a workaround to avoid an endless loop.  See #15873/#15984 *)
       match Sentence.find buffer iter with
       | None -> ()
       | Some (start, stop) ->
@@ -640,7 +682,7 @@ object(self)
             stop#equal (buffer#get_iter_at_mark s.stop)) topstack
         then begin
           Queue.push (`Skip (start, stop)) queue;
-          loop n stop
+          loop n stop start#offset
         end else begin
           buffer#apply_tag Tags.Script.to_process ~start ~stop;
           let sentence =
@@ -649,10 +691,10 @@ object(self)
               ~stop:(`MARK (buffer#create_mark stop))
               [] in
           Queue.push (`Sentence sentence) queue;
-          if not stop#is_end then loop (succ n) stop
+          if start#offset <> prev_start && not stop#is_end then loop (succ n) stop start#offset
         end
     in
-    loop 0 self#get_start_of_input
+    loop 0 self#get_start_of_input (-1)
 
   method private discard_command_queue queue =
     while not (Queue.is_empty queue) do
@@ -704,9 +746,15 @@ object(self)
             add_flag sentence `PROCESSING;
             Doc.push document sentence;
             let start, _, phrase = self#get_sentence sentence in
-            let bp = c2b buffer start#offset in
+            (* script before last_offsets is not editable because Coq is processing *)
+            let s_offs = if start#offset > (snd last_offsets) then last_offsets else (0,0) in
+            let bp = c2b buffer s_offs start#offset in
+            sentence.byte_start <- bp;
             let line_nb = start#line + 1 in
-            let bol_pos = c2b buffer start#line_offset in
+            let bol_uni = start#offset - start#line_offset in
+            let bol_byte = c2b buffer s_offs bol_uni in
+            let bol_pos = bp - bol_byte in
+            last_offsets <- (bol_byte, bol_uni);
             let coq_query = Coq.add ((((phrase,edit_id),(tip,verbose)),bp),(line_nb,bol_pos)) in
             let handle_answer = function
               | Good (id, Util.Inl (* NewTip *) ()) ->
@@ -760,7 +808,7 @@ object(self)
            | Some loc ->
              let (iter, _, phrase) = self#get_sentence s in
              let (start, _) = Loc.unloc loc in
-             iter#forward_chars (b2c buffer start)
+             buffer#get_iter (`OFFSET (b2c phrase start))
          end in iter#line + 1, msg
       | _ -> assert false in
     List.rev
@@ -871,6 +919,10 @@ object(self)
     in
       undo to_id unfocus_needed)
 
+  method backtrack_to_begin =
+    let until (id : Doc.id option) _ stop = (id = (Some (Stateid.of_int 1))) in
+    fun () -> self#backtrack_to_id ~move_insert:false (self#find_id until)
+
   method private backtrack_until ?move_insert until =
     self#backtrack_to_id ?move_insert (self#find_id until)
 
