@@ -70,12 +70,6 @@ let on_current_term f =
   | None -> ()
   | Some t -> ignore (f t)
 
-let on_current_term_val default f =
-  let term = try Some notebook#current_term with Invalid_argument _ -> None in
-  match term with
-  | None -> default
-  | Some t -> f t
-
 let cb_on_current_term f _ = on_current_term f
 
 (** Nota: using && here has the advantage of working both under win32 and unix.
@@ -109,6 +103,7 @@ let make_coqtop_args fname =
         with
         | None -> "", base_args
         | Some proj ->
+          Minilib.log ("About to read project file " ^ proj);
           let warning_fn x = Feedback.msg_warning Pp.(str x) in
           proj, coqtop_args_from_project (read_project_file ~warning_fn proj) @ base_args
   in
@@ -203,7 +198,9 @@ let create_session f =
   sn.coqops#set_forward_set_goals_of_dbg_session
     (fun msg ->
       sn.coqops#set_debug_goal msg;
+      sn.last_db_goals <- msg;
       let osn = (find_secondary_sn sn) in
+      osn.coqops#set_debug_goal msg;
       osn.last_db_goals <- msg);
   sn.coqops#set_forward_init_db (fun () -> !forward_init_db sn);
   let _ = set_drag sn.script#drag in
@@ -445,8 +442,12 @@ let revert_all ?parent _ =
       end)
     notebook#pages
 
+let win_interrupt = ref false
+
 let quit ?parent _ =
-  if not !FileAux.in_quit_dialog then
+  if !win_interrupt then
+    win_interrupt := false
+  else if not !FileAux.in_quit_dialog then
     try FileAux.check_quit ?parent saveall; exit 0
     with FileAux.DontQuit -> ()
 
@@ -845,6 +846,7 @@ module Nav = struct
       Coq.reset_coqtop sn.coqtop (* calls init_bpts *)
     end
   let interrupt _ =  (* terminate computation *)
+    if Sys.os_type = "Win32" then File.win_interrupt := true;
     Minilib.log "User interrupt received";
     if not (resume_debugger Interface.Interrupt) && notebook#pages <> [] then begin
       let osn = (find_db_sn ()) in
@@ -1182,13 +1184,6 @@ let apply_unicode_binding =
   cb_on_current_term (fun t ->
     t.script#apply_unicode_binding())
 
-let mem_usage =
-  cb_on_current_term (fun t ->
-    let open Gc in
-    let stats = Gc.stat () in
-    Printf.printf "live_words = %d heap_words = %d top_heap_words = %d\n%!"
-      stats.live_words stats.heap_words stats.top_heap_words)
-
 let coqtop_arguments sn =
   init_bpts sn;
   let dialog = GWindow.dialog ~title:"Coqtop arguments" () in
@@ -1294,17 +1289,6 @@ let toggle_items menu_name l =
 
 let no_under = Util.String.map (fun x -> if x = '_' then '-' else x)
 
-(* workaround: GTKSourceView3 ignores editable *)
-
-(* allowed in script panel, if editable, and elsewhere *)
-let can_paste () =
-  on_current_term_val true (fun t ->
-    (t.script#is_focus && t.script#editable2) || not t.script#is_focus)
-
-(* template insert only allowed in script panel *)
-let can_template () =
-  on_current_term_val false (fun t -> t.script#is_focus && t.script#editable2)
-
 (** Create alphabetical menu items with elements in sub-items.
     [l] is a list of lists, one per initial letter *)
 
@@ -1327,9 +1311,7 @@ let alpha_items menu_name item_name l =
       Buffer.contents buf
     in
     let callback _ =
-      on_current_term (fun sn ->
-        let text' = if can_template () then text' else "" in
-                      sn.buffer#insert_interactive text')
+      on_current_term (fun sn -> sn.buffer#insert_interactive text')
     in
     item (item_name^" "^(no_under text)) ~label:text ~callback menu_name
   in
@@ -1356,17 +1338,16 @@ let template_item (text, offset, len, key) =
   let label = "_"^name^" __" in
   let negoffset = String.length text - offset - len in
   let callback sn =
-    if can_paste () then
-      let b = sn.buffer in
-      if b#insert_interactive text then begin
-        let iter = b#get_iter_at_mark `INSERT in
-        ignore (iter#nocopy#backward_chars negoffset);
-        b#move_mark `INSERT ~where:iter;
-        ignore (iter#nocopy#backward_chars len);
-        b#move_mark `SEL_BOUND ~where:iter;
-      end
-    in
-    item name ~label ~callback:(cb_on_current_term callback) ~accel:(modifier^key)
+    let b = sn.buffer in
+    if b#insert_interactive text then begin
+      let iter = b#get_iter_at_mark `INSERT in
+      ignore (iter#nocopy#backward_chars negoffset);
+      b#move_mark `INSERT ~where:iter;
+      ignore (iter#nocopy#backward_chars len);
+      b#move_mark `SEL_BOUND ~where:iter;
+    end
+  in
+  item name ~label ~callback:(cb_on_current_term callback) ~accel:(modifier^key)
 
 (** Create menu items for pairs (query, shortcut key). *)
 let user_queries_items menu_name item_name l =
@@ -1386,6 +1367,16 @@ let emit_to_focus window sgn =
 (** {2 Creation of the main coqide window } *)
 
 let build_ui () =
+  (* issue 12779 *)
+  GtkData.StyleContext.add_provider_for_screen
+    (Gdk.Screen.default ())
+    begin
+      let provider = GObj.css_provider () in
+      provider#load_from_data "button { box-shadow: none }";
+      provider#as_css_provider
+    end
+    GtkData.StyleContext.ProviderPriority.application;
+
   let w = GWindow.window
     ~wmclass:("CoqIDE","CoqIDE")
     ~width:window_width#get ~height:window_height#get
@@ -1447,19 +1438,15 @@ let build_ui () =
   menu edit_menu [
     item "Edit" ~label:"_Edit";
     item "Undo" ~accel:"<Primary>u" ~stock:`UNDO
-      ~callback:(cb_on_current_term (fun t -> if can_paste () then
-          t.script#undo ()));
+      ~callback:(cb_on_current_term (fun t -> t.script#undo ()));
     item "Redo" ~stock:`REDO
-      ~callback:(cb_on_current_term (fun t -> if can_paste () then
-          t.script#redo ()));
+      ~callback:(cb_on_current_term (fun t -> t.script#redo ()));
     item "Cut" ~stock:`CUT
-      ~callback:(fun _ -> if can_paste () then
-          emit_to_focus w GtkText.View.S.cut_clipboard);
+      ~callback:(fun _ -> emit_to_focus w GtkText.View.S.cut_clipboard);
     item "Copy" ~stock:`COPY
       ~callback:(fun _ -> emit_to_focus w GtkText.View.S.copy_clipboard);
     item "Paste" ~stock:`PASTE
-      ~callback:(fun _ -> if can_paste () then
-          emit_to_focus w GtkText.View.S.paste_clipboard);
+      ~callback:(fun _ -> emit_to_focus w GtkText.View.S.paste_clipboard);
     item "Find" ~stock:`FIND ~label:"Find / Replace"
       ~callback:(cb_on_current_term (fun t -> t.finder#show ()));
     item "Find Next" ~label:"Find _Next" ~stock:`GO_DOWN ~accel:"F3"
@@ -1589,11 +1576,9 @@ let build_ui () =
   menu tools_menu [
     item "Tools" ~label:"_Tools";
     item "Comment" ~label:"_Comment" ~accel:"<CTRL>D"
-      ~callback:(cb_on_current_term (fun t -> if can_paste () then
-          t.script#comment ()));
+      ~callback:(cb_on_current_term (fun t -> t.script#comment ()));
     item "Uncomment" ~label:"_Uncomment" ~accel:"<CTRL><SHIFT>D"
-      ~callback:(cb_on_current_term (fun t -> if can_paste () then
-          t.script#uncomment ()));
+      ~callback:(cb_on_current_term (fun t -> t.script#uncomment ()));
     item "Coqtop arguments" ~label:"Coqtop _arguments"
       ~callback:MiscMenu.coqtop_arguments;
     item "LaTeX-to-unicode" ~label:"_LaTeX-to-unicode" ~accel:"<Shift>space"
@@ -1642,8 +1627,6 @@ let build_ui () =
       ~callback:(fun _ -> on_current_term (fun sn ->
          sn.messages#default_route#clear;
          sn.messages#default_route#add_string (MicroPG.get_documentation ())));
-    item "Memory usage" ~label:"Memory usage"  (* temporary, for debugging *)
-      ~callback:MiscMenu.mem_usage;
     item "About Coq" ~label:"_About" ~stock:`ABOUT
       ~callback:MiscMenu.about
   ];
@@ -1802,6 +1785,20 @@ let main files =
     in the path. Note that the -coqtop option to coqide overrides
     this default coqtop path *)
 
+let coqide_specific_usage = Boot.Usage.{
+  executable_name = "coqide";
+  extra_args = "";
+  extra_options = "\n\
+CoqIDE specific options:\
+\n  -f _CoqProjectFile         set _CoqProject file to _CoqProjectFile\
+\n  -unicode-bindings f1 .. f2 load files f1..f2 with extra unicode bindings\
+\n  -coqtop dir                look for coqidetop in dir\
+\n  -coqtop-flags              extra flags for the coqtop subprocess\
+\n  -debug                     enable debug mode\
+\n  -xml-debug                 enable debug mode and print XML messages to/from CoqIDE\
+\n"
+}
+
 let read_coqide_args argv =
   let set_debug () =
     CDebug.set_debug_all true;
@@ -1815,17 +1812,17 @@ let read_coqide_args argv =
     |"-coqtop" :: prog :: args ->
       if coqtop = None then filter_coqtop (Some prog) project_files bindings_files out args
       else (output_string stderr "Error: multiple -coqtop options"; exit 1)
+    |"-coqtop" :: [] ->
+      output_string stderr "Error: missing argument after -coqtop"; exit 1
     |"-f" :: file :: args ->
       if project_files <> None then
         (output_string stderr "Error: multiple -f options"; exit 1);
       let d = CUnix.canonical_path_name (Filename.dirname file) in
       let warning_fn x = Format.eprintf "%s@\n%!" x in
       let p = CoqProject_file.read_project_file ~warning_fn file in
-      filter_coqtop coqtop (Some (d,p)) out bindings_files args
+      filter_coqtop coqtop (Some (d,p)) bindings_files out args
     |"-f" :: [] ->
       output_string stderr "Error: missing project file name"; exit 1
-    |"-coqtop" :: [] ->
-      output_string stderr "Error: missing argument after -coqtop"; exit 1
     |"-debug"::args ->
       set_debug ();
       filter_coqtop coqtop project_files bindings_files out args
@@ -1837,12 +1834,17 @@ let read_coqide_args argv =
       Coq.ideslave_coqtop_flags := Some flags;
       filter_coqtop coqtop project_files bindings_files out args
     | ("-v" | "--version") :: _ ->
+      (* This does the same thing as Usage.version () but printed differently *)
       Printf.printf "CoqIDE, version %s\n" Coq_config.version;
+      (* Unlike coqtop we don't accumulate queries so we exit immediately *)
       exit 0
-    |arg::args when out = [] && CString.is_prefix "-psn_" arg ->
+    | ("-h"|"-H"|"-?"|"-help"|"--help") :: _ ->
+      Boot.Usage.print_usage stderr coqide_specific_usage;
+      exit 0
+    | arg::args when out = [] && CString.is_prefix "-psn_" arg ->
       (* argument added by MacOS during .app launch *)
       filter_coqtop coqtop project_files bindings_files out args
-    |arg::args -> filter_coqtop coqtop project_files bindings_files (arg::out) args
+    | arg::args -> filter_coqtop coqtop project_files bindings_files (arg::out) args
     |[] -> (coqtop,project_files,bindings_files,List.rev out)
   in
   let coqtop,project_files,bindings_files,argv = filter_coqtop None None [] [] argv in
