@@ -96,13 +96,13 @@ val intern_context : env -> bound_univs:UnivNames.universe_binders ->
 (** Main interpretation functions, using type class inference,
     expecting evars and pending problems to be all resolved *)
 
-val interp_constr : ?expected_type:typing_constraint -> env -> evar_map -> ?impls:internalization_env ->
+val interp_constr : ?flags:inference_flags -> ?expected_type:typing_constraint -> env -> evar_map -> ?impls:internalization_env ->
   constr_expr -> constr Evd.in_evar_universe_context
 
-val interp_casted_constr : env -> evar_map -> ?impls:internalization_env ->
+val interp_casted_constr : ?flags:inference_flags -> env -> evar_map -> ?impls:internalization_env ->
   constr_expr -> types -> constr Evd.in_evar_universe_context
 
-val interp_type : env -> evar_map -> ?impls:internalization_env ->
+val interp_type : ?flags:inference_flags -> env -> evar_map -> ?impls:internalization_env ->
   constr_expr -> types Evd.in_evar_universe_context
 
 (** Main interpretation function expecting all postponed problems to
@@ -147,15 +147,15 @@ val interp_constr_pattern :
   env -> evar_map -> ?expected_type:typing_constraint ->
     constr_pattern_expr -> constr_pattern
 
-(** Returns None if it's a syndef not bound to a name, raises an error
+(** Returns None if it's an abbreviation not bound to a name, raises an error
     if not existing *)
 val intern_reference : qualid -> GlobRef.t option
 
-(** Returns None if not a reference or a syndef not bound to a name *)
+(** Returns None if not a reference or a abbrev not bound to a name *)
 val intern_name_alias :
   constr_expr -> (GlobRef.t * Glob_term.glob_level list option) option
 
-(** Expands abbreviations (syndef); raise an error if not existing *)
+(** Expands abbreviations; raise an error if not existing *)
 val interp_reference : ltac_sign -> qualid -> glob_constr
 
 (** Interpret binders *)
@@ -211,7 +211,10 @@ val check_duplicate : ?loc:Loc.t -> (qualid * constr_expr) list -> unit
 (** Check that a list of record field definitions doesn't contain
     duplicates. *)
 
-val interp_known_level : Evd.evar_map -> sort_name_expr -> Univ.Level.t
+val interp_univ_constraint
+  : Evd.evar_map
+  -> sort_name_expr * Univ.constraint_type * sort_name_expr
+  -> Univ.univ_constraint
 
 (** Local universe and constraint declarations. *)
 val interp_univ_decl : Environ.env -> universe_decl_expr ->
