@@ -41,6 +41,17 @@ let pop_output () = glob_output := match !glob_output with
 let pause () = push_output NoGlob
 let continue = pop_output
 
+let with_glob_output g f () =
+  push_output g;
+  try
+    let res = f () in
+    pop_output ();
+    res
+  with reraise ->
+    let reraise = Exninfo.capture reraise in
+    pop_output ();
+    Exninfo.iraise reraise
+
 let dump () = get_output () <> NoGlob
 
 let dump_string s =
@@ -248,9 +259,9 @@ let mp_of_kn kn =
 
 let add_glob_kn ?loc kn =
   if dump () then
-    let sp = Nametab.path_of_syndef kn in
+    let sp = Nametab.path_of_abbreviation kn in
     let lib_dp = Lib.dp_of_mp (mp_of_kn kn) in
-    add_glob_gen ?loc sp lib_dp "syndef"
+    add_glob_gen ?loc sp lib_dp "abbrev"
 
 let dump_def ?loc ty secpath id = Option.iter (fun loc ->
   if get_output () = Feedback then
