@@ -32,6 +32,15 @@ let anomaly ?loc ?info ?label pp =
 
 exception UserError of Pp.t (* User errors *)
 
+(* We register Coq Errors with the global printer here as they will be
+   printed if Dynlink.loadfile raises UserError in the module
+   intializers, and in some other cases. We should make this more
+   principled. *)
+let _ = Printexc.register_printer (function
+    | UserError msg ->
+      Some (Format.asprintf "@[Coq Error: %a@]" Pp.pp_with msg)
+    | _ -> None)
+
 let user_err ?loc ?info strm =
   let info = Option.default Exninfo.null info in
   let info = Option.cata (Loc.add_loc info) info loc in
