@@ -35,7 +35,7 @@ let environment_until dir_opt =
     | [] -> []
     | d :: l ->
       let meb =
-        Modops.destr_nofunctor (Global.lookup_module (MPfile d)).mod_type
+        Modops.destr_nofunctor (MPfile d) (Global.lookup_module (MPfile d)).mod_type
       in
       match dir_opt with
       | Some d' when DirPath.equal d d' -> [MPfile d, meb]
@@ -159,11 +159,14 @@ let factor_fix env sg l cb msb =
 
 let expand_mexpr env mpo me =
   let inl = Some (Flags.get_inline_level()) in
-  Mod_typing.translate_mse env mpo inl me
+  let state = ((Environ.universes env, Univ.Constraints.empty), Reductionops.inferred_universes) in
+  let sign, expr, delta, (_, cst) = Mod_typing.translate_mse state env mpo inl me in
+  sign, expr, delta, cst
 
 let expand_modtype env mp me =
   let inl = Some (Flags.get_inline_level()) in
-  let mtb, _cst = Mod_typing.translate_modtype env mp inl ([],me) in
+  let state = ((Environ.universes env, Univ.Constraints.empty), Reductionops.inferred_universes) in
+  let mtb, _cst = Mod_typing.translate_modtype state env mp inl ([],me) in
   mtb
 
 let no_delta = Mod_subst.empty_delta_resolver
@@ -179,7 +182,7 @@ let flatten_modtype env mp me_alg struc_opt =
 *)
 
 let env_for_mtb_with_def env mp me reso idl =
-  let struc = Modops.destr_nofunctor me in
+  let struc = Modops.destr_nofunctor mp me in
   let l = Label.of_id (List.hd idl) in
   let spot = function (l',SFBconst _) -> Label.equal l l' | _ -> false in
   let before = fst (List.split_when spot struc) in
