@@ -20,6 +20,7 @@ open Proofview.Notations
 let constr_flags =
   let open Pretyping in
   {
+    use_coercions = true;
     use_typeclasses = Pretyping.UseTC;
     solve_unification_constraints = true;
     fail_evar = true;
@@ -31,6 +32,7 @@ let constr_flags =
 let open_constr_no_classes_flags =
   let open Pretyping in
   {
+  use_coercions = true;
   use_typeclasses = Pretyping.NoUseTC;
   solve_unification_constraints = true;
   fail_evar = false;
@@ -140,6 +142,9 @@ let err_notfound =
 let err_matchfailure =
   Tac2interp.LtacError (coq_core "Match_failure", [||])
 
+let err_division_by_zero =
+  Tac2interp.LtacError (coq_core "Division_by_zero", [||])
+
 (** Helper functions *)
 
 let thaw f = Tac2ffi.apply f [v_unit]
@@ -162,7 +167,7 @@ let fail ?(info = Exninfo.null) e =
   Proofview.tclZERO ~info e
 
 let return x = Proofview.tclUNIT x
-let pname s = { mltac_plugin = "ltac2"; mltac_tactic = s }
+let pname s = { mltac_plugin = "coq-core.plugins.ltac2"; mltac_tactic = s }
 
 let wrap f =
   return () >>= fun () -> return (f ())
@@ -386,6 +391,10 @@ let () = define2 "int_equal" int int begin fun m n ->
   return (Value.of_bool (m == n))
 end
 
+let unop n f = define1 n int begin fun m ->
+  return (Value.of_int (f m))
+end
+
 let binop n f = define2 n int int begin fun m n ->
   return (Value.of_int (f m n))
 end
@@ -394,10 +403,24 @@ let () = binop "int_compare" Int.compare
 let () = binop "int_add" (+)
 let () = binop "int_sub" (-)
 let () = binop "int_mul" ( * )
-
-let () = define1 "int_neg" int begin fun m ->
-  return (Value.of_int (~- m))
+let () = define2 "int_div" int int begin fun m n ->
+  if n == 0 then throw err_division_by_zero
+  else return (Value.of_int (m / n))
 end
+let () = define2 "int_mod" int int begin fun m n ->
+  if n == 0 then throw err_division_by_zero
+  else return (Value.of_int (m mod n))
+end
+let () = unop "int_neg" (~-)
+let () = unop "int_abs" abs
+
+let () = binop "int_asr" (asr)
+let () = binop "int_lsl" (lsl)
+let () = binop "int_lsr" (lsr)
+let () = binop "int_land" (land)
+let () = binop "int_lor" (lor)
+let () = binop "int_lxor" (lxor)
+let () = unop "int_lnot" lnot
 
 (** Char *)
 
@@ -435,7 +458,7 @@ end
 (** constr -> constr *)
 let () = define1 "constr_type" constr begin fun c ->
   let get_type env sigma =
-  Proofview.V82.wrap_exceptions begin fun () ->
+  Proofview.wrap_exceptions begin fun () ->
     let (sigma, t) = Typing.type_of env sigma c in
     let t = Value.of_constr t in
     Proofview.Unsafe.tclEVARS sigma <*> Proofview.tclUNIT t
@@ -719,7 +742,7 @@ let () = define1 "constr_pretype" (repr_ext val_preterm) begin fun c ->
   let open Pretyping in
   let open Ltac_pretype in
   let pretype env sigma =
-  Proofview.V82.wrap_exceptions begin fun () ->
+  Proofview.wrap_exceptions begin fun () ->
     (* For now there are no primitives to create preterms with a non-empty
        closure. I do not know whether [closed_glob_constr] is really the type
        we want but it does not hurt in the meantime. *)
@@ -763,10 +786,10 @@ end
 
 (** Patterns *)
 
-let empty_context = EConstr.mkMeta Constr_matching.special_meta
+let empty_context = Constr_matching.empty_context
 
 let () = define0 "pattern_empty_context" begin
-  return (Value.of_constr empty_context)
+  return (Value.of_ext val_matching_context empty_context)
 end
 
 let () = define2 "pattern_matches" pattern constr begin fun pat c ->
@@ -792,10 +815,11 @@ let () = define2 "pattern_matches_subterm" pattern constr begin fun pat c ->
   | IStream.Cons ({ m_sub = (_, sub); m_ctx }, s) ->
     let ans = Id.Map.bindings sub in
     let of_pair (id, c) = Value.of_tuple [| Value.of_ident id; Value.of_constr c |] in
-    let ans = Value.of_tuple [| Value.of_constr (Lazy.force m_ctx); Value.of_list of_pair ans |] in
+    let ans = Value.of_tuple [| Value.of_ext val_matching_context m_ctx; Value.of_list of_pair ans |] in
     Proofview.tclOR (return ans) (fun _ -> of_ans s)
   in
   pf_apply begin fun env sigma ->
+    let pat = Constr_matching.instantiate_pattern env sigma Id.Map.empty pat in
     let ans = Constr_matching.match_subterm env sigma (Id.Set.empty,pat) c in
     of_ans ans
   end
@@ -824,10 +848,11 @@ let () = define2 "pattern_matches_subterm_vect" pattern constr begin fun pat c -
   | IStream.Cons ({ m_sub = (_, sub); m_ctx }, s) ->
     let ans = Id.Map.bindings sub in
     let ans = Array.map_of_list snd ans in
-    let ans = Value.of_tuple [| Value.of_constr (Lazy.force m_ctx); Value.of_array Value.of_constr ans |] in
+    let ans = Value.of_tuple [| Value.of_ext val_matching_context m_ctx; Value.of_array Value.of_constr ans |] in
     Proofview.tclOR (return ans) (fun _ -> of_ans s)
   in
   pf_apply begin fun env sigma ->
+    let pat = Constr_matching.instantiate_pattern env sigma Id.Map.empty pat in
     let ans = Constr_matching.match_subterm env sigma (Id.Set.empty,pat) c in
     of_ans ans
   end
@@ -842,7 +867,7 @@ let () = define3 "pattern_matches_goal" bool (list (pair bool pattern)) (pair bo
   let mk_pattern (b, pat) = if b then Tac2match.MatchPattern pat else Tac2match.MatchContext pat in
   let r = (List.map mk_pattern hp, mk_pattern cp) in
   Tac2match.match_goal env sigma concl ~rev r >>= fun (hyps, ctx, subst) ->
-    let of_ctxopt ctx = Value.of_constr (Option.default empty_context ctx) in
+    let of_ctxopt ctx = Value.of_ext val_matching_context (Option.default empty_context ctx) in
     let hids = Value.of_array Value.of_ident (Array.map_of_list fst hyps) in
     let hctx = Value.of_array of_ctxopt (Array.map_of_list snd hyps) in
     let subs = Value.of_array Value.of_constr (Array.map_of_list snd (Id.Map.bindings subst)) in
@@ -852,11 +877,9 @@ let () = define3 "pattern_matches_goal" bool (list (pair bool pattern)) (pair bo
   end
 end
 
-let () = define2 "pattern_instantiate" constr constr begin fun ctx c ->
-  let ctx = EConstr.Unsafe.to_constr ctx in
-  let c = EConstr.Unsafe.to_constr c in
-  let ans = Termops.subst_meta [Constr_matching.special_meta, c] ctx in
-  return (Value.of_constr (EConstr.of_constr ans))
+let () = define2 "pattern_instantiate" (repr_ext val_matching_context) constr begin fun ctx c ->
+  let ans = Constr_matching.instantiate_context ctx c in
+  return (Value.of_constr ans)
 end
 
 (** Error *)
@@ -1556,7 +1579,7 @@ let () =
   in
   let () = Geninterp.register_interp0 wit_ltac2_val interp_fun in
   define1 "ltac1_lambda" valexpr begin fun f ->
-    let body = Tacexpr.TacGeneric (Some "ltac2", in_gen (glbwit wit_ltac2_val) ()) in
+    let body = Tacexpr.TacGeneric (Some "coq-core.plugins.ltac2", in_gen (glbwit wit_ltac2_val) ()) in
     let clos = CAst.make (Tacexpr.TacFun ([Name arg_id], CAst.make (Tacexpr.TacArg body))) in
     let f = Geninterp.Val.inject (Geninterp.Val.Base typ_ltac2) f in
     let lfun = Id.Map.singleton tac_id f in
@@ -1567,7 +1590,7 @@ let () =
 let ltac2_eval =
   let open Ltac_plugin in
   let ml_name = {
-    Tacexpr.mltac_plugin = "ltac2";
+    Tacexpr.mltac_plugin = "coq-core.plugins.ltac2";
     mltac_tactic = "ltac2_eval";
   } in
   let eval_fun args ist = match args with
