@@ -131,7 +131,7 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
       begin match ESorts.kind !evdref s with
       | Type u ->
          (* TODO: check if max(l,u) is not ok as well *)
-        (match Univ.universe_level u with
+        (match Univ.Universe.level u with
         | None -> refresh_sort status ~direction s
         | Some l ->
            (match Evd.universe_rigidity !evdref l with
@@ -199,9 +199,11 @@ let refresh_universes ?(status=univ_rigid) ?(onlyalg=false) ?(refreshset=false)
     else refresh_term_evars ~onevars:false ~top:true t
   in !evdref, t'
 
-let get_type_of_refresh ?(polyprop=true) ?(lax=false) env sigma c =
-  let ty = Retyping.get_type_of ~polyprop ~lax env sigma c in
-    refresh_universes (Some false) env sigma ty
+let get_type_of_refresh  ?(polyprop=true) ?(lax=false) env evars t =
+  let tty = Retyping.get_type_of env evars t in
+  let evars', tty = refresh_universes ~onlyalg:true
+    ~status:(Evd.UnivFlexible false) (Some false) env evars tty in
+  evars', tty
 
 let add_conv_oriented_pb ?(tail=true) (pbty,env,t1,t2) evd =
   match pbty with
@@ -591,14 +593,15 @@ let alias_distinct l =
   in
   check (Int.Set.empty, Id.Set.empty) l
 
-let get_actual_deps env evd aliases l t =
-  if occur_meta evd t then
-    (* The instance of a meta can virtually contains any variable of the context *)
-    l
+let distinct_actual_deps env evd aliases l t =
+  (* If the aliases are already unique, any subset will also be. *)
+  if alias_distinct l then true
+  (* The instance of a meta can virtually contains any variable of the context *)
+  else if occur_meta evd t then false
   else
     (* Probably strong restrictions coming from t being evar-closed *)
     let (fv_rels,fv_ids,_,_) = free_vars_and_rels_up_alias_expansion env evd aliases t in
-    List.filter (function
+    alias_distinct @@ List.filter (function
       | VarAlias id -> Id.Set.mem id fv_ids
       | RelAlias n -> Int.Set.mem n fv_rels
     ) l
@@ -619,7 +622,7 @@ let remove_instance_local_defs evd evk args =
 let find_unification_pattern_args env evd l t =
   let aliases = make_alias_map env evd in
   match expand_and_check_vars evd aliases l with
-  | Some l as x when alias_distinct (get_actual_deps env evd aliases l t) -> x
+  | Some l as x when distinct_actual_deps env evd aliases l t -> x
   | _ -> None
 
 let is_unification_pattern_meta env evd nb m l t =
@@ -1065,7 +1068,7 @@ let extract_unique_projection = function
 let extract_candidates sols =
   try
     UpdateWith
-      (List.map (function (id,ProjectVar) -> mkVar id | _ -> raise Exit) sols)
+      (List.map (function (id,ProjectVar) -> mkVar id | _ -> raise_notrace Exit) sols)
   with Exit ->
     NoUpdate
 
@@ -1401,14 +1404,13 @@ let solve_evar_evar ?(force=false) f unify flags env evd pbty (evk1,args1 as ev1
       let concl2 = EConstr.Unsafe.to_constr evi2.evar_concl in
       let ctx2, j = Reduction.dest_arity evi2env concl2 in
       let ctx2 = List.map (fun c -> map_rel_decl EConstr.of_constr c) ctx2 in
-      let ui, uj = univ_of_sort i, univ_of_sort j in
-        if i == j || Evd.check_eq evd ui uj
+        if i == j || Evd.check_eq evd i j
         then (* Shortcut, i = j *)
           evd
-        else if Evd.check_leq evd ui uj then
+        else if Evd.check_leq evd i j then
           let t2 = it_mkProd_or_LetIn (mkSort i) ctx2 in
           downcast evk2 t2 evd
-        else if Evd.check_leq evd uj ui then
+        else if Evd.check_leq evd j i then
           let t1 = it_mkProd_or_LetIn (mkSort j) ctx1 in
           downcast evk1 t1 evd
         else
