@@ -155,7 +155,9 @@ let head_of_constr_reference sigma c = match EConstr.kind sigma c with
   | Var id -> GlobRef.VarRef id
   | _ -> anomaly (Pp.str "Not a rigid reference.")
 
-let pattern_of_constr env sigma t =
+let pattern_of_constr ~broken env sigma t =
+  let t = EConstr.Unsafe.to_constr t in
+  let kind = if broken then Constr.kind else fun c -> EConstr.kind_upto sigma c in
   let rec pattern_of_constr env t =
   let open Context.Rel.Declaration in
   match kind t with
@@ -238,6 +240,9 @@ let pattern_of_constr env sigma t =
     in
   pattern_of_constr env t
 
+let legacy_bad_pattern_of_constr env sigma c : constr_pattern = pattern_of_constr ~broken:true env sigma c
+let pattern_of_constr env sigma c : constr_pattern = pattern_of_constr ~broken:false env sigma c
+
 (* To process patterns, we need a translation without typing at all. *)
 
 let map_pattern_with_binders g f l = function
@@ -264,42 +269,6 @@ let map_pattern_with_binders g f l = function
   | (PVar _ | PEvar _ | PRel _ | PRef _  | PSort _  | PMeta _ | PInt _
      | PFloat _ as x) -> x
 
-let error_instantiate_pattern id l =
-  let is = match l with
-  | [_] -> "is"
-  | _ -> "are"
-  in
-  user_err  (str "Cannot substitute the term bound to " ++ Id.print id
-    ++ strbrk " in pattern because the term refers to " ++ pr_enum Id.print l
-    ++ strbrk " which " ++ str is ++ strbrk " not bound in the pattern.")
-
-let instantiate_pattern env sigma lvar c =
-  let open EConstr in
-  let open Vars in
-  let rec aux vars = function
-  | PVar id as x ->
-      (try
-        let ctx,c = Id.Map.find id lvar in
-        try
-          let inst =
-            List.map
-              (fun id -> mkRel (List.index Name.equal (Name id) vars))
-              ctx
-          in
-          let c = substl inst c in
-          (* FIXME: Stupid workaround to pattern_of_constr being evar sensitive *)
-          let c = Evarutil.nf_evar sigma c in
-          pattern_of_constr env sigma (EConstr.Unsafe.to_constr c)
-        with Not_found (* List.index failed *) ->
-          let vars =
-            List.map_filter (function Name id -> Some id | _ -> None) vars in
-          error_instantiate_pattern id (List.subtract Id.equal ctx vars)
-       with Not_found (* Map.find failed *) ->
-         x)
-  | c ->
-      map_pattern_with_binders (fun id vars -> id::vars) aux vars c in
-  aux [] c
-
 let rec liftn_pattern k n = function
   | PRel i as x -> if i >= n then PRel (i+k) else x
   | c -> map_pattern_with_binders (fun _ -> succ) (liftn_pattern k) n c
@@ -313,7 +282,7 @@ let rec subst_pattern env sigma subst pat =
     if ref' == ref then pat else (match t with
         | None -> PRef ref'
         | Some t ->
-          pattern_of_constr env sigma t.Univ.univ_abstracted_value)
+          pattern_of_constr env sigma (EConstr.of_constr t.Univ.univ_abstracted_value))
   | PVar _
   | PEvar _
   | PRel _
