@@ -62,13 +62,15 @@ let pr_econstr_n_env ?lax ?inctx ?scope env sigma n t =
   pr_constr_expr_n env sigma n (extern_constr ?lax ?inctx ?scope env sigma t)
 let pr_econstr_env ?lax ?inctx ?scope env sigma t =
   pr_constr_expr env sigma (extern_constr ?lax ?inctx ?scope env sigma t)
-let pr_leconstr_env = Proof_diffs.pr_leconstr_env
+let pr_leconstr_env ?lax ?inctx ?scope env sigma t =
+  Ppconstr.pr_lconstr_expr env sigma (extern_constr ?lax ?inctx ?scope env sigma t)
 
 let pr_constr_n_env ?lax ?inctx ?scope env sigma n c =
   pr_econstr_n_env ?lax ?inctx ?scope env sigma n (EConstr.of_constr c)
 let pr_constr_env ?lax ?inctx ?scope env sigma c =
   pr_econstr_env ?lax ?inctx ?scope env sigma (EConstr.of_constr c)
-let pr_lconstr_env = Proof_diffs.pr_lconstr_env
+let pr_lconstr_env ?lax ?inctx ?scope env sigma c =
+  pr_leconstr_env ?lax ?inctx ?scope env sigma (EConstr.of_constr c)
 
 let pr_open_lconstr_env ?lax ?inctx ?scope env sigma (_,c) =
   pr_leconstr_env ?lax ?inctx ?scope env sigma c
@@ -87,7 +89,8 @@ let pr_lconstr_under_binders_env = pr_constr_under_binders_env_gen pr_leconstr_e
 
 let pr_etype_env ?lax ?goal_concl_style env sigma t =
   pr_constr_expr env sigma (extern_type ?lax ?goal_concl_style env sigma t)
-let pr_letype_env = Proof_diffs.pr_letype_env
+let pr_letype_env ?lax ?goal_concl_style env sigma ?impargs t =
+  pr_lconstr_expr env sigma (extern_type ?lax ?goal_concl_style env sigma ?impargs t)
 
 let pr_type_env ?lax ?goal_concl_style env sigma c =
   pr_etype_env ?lax ?goal_concl_style env sigma (EConstr.of_constr c)
@@ -465,14 +468,22 @@ let goal_repr sigma g =
  og_s has goal+sigma on the previous proof step for diffs
  g_s has goal+sigma on the current proof step
  *)
-let pr_goal ?(diffs=false) ?og_s g_s =
-  let g = sig_it g_s in
-  let sigma = Tacmach.Old.project g_s in
-  let env, concl = goal_repr sigma g in
-  let goal =
-    if diffs then
-      Proof_diffs.diff_goal ?og_s g sigma
-    else
+let pr_goal ?diffs sigma g =
+  let goal = match diffs with
+  | Some og_s ->
+    let g = Proof_diffs.make_goal (Global.env ()) sigma g in
+    let (hyps_pp_list, concl_pp) = Proof_diffs.diff_goal ?og_s g in
+    let hyp_list_to_pp hyps =
+      match hyps with
+      | h :: tl -> List.fold_left (fun x y -> x ++ cut () ++ y) h tl
+      | [] -> mt ()
+    in
+    v 0 (
+      (hyp_list_to_pp hyps_pp_list) ++ cut () ++
+      str "============================" ++ cut () ++
+      concl_pp)
+  | None ->
+    let env, concl = goal_repr sigma g in
       pr_context_of env sigma ++ cut () ++
         str "============================" ++ cut ()  ++
         hov 0 (pr_letype_env ~goal_concl_style:true env sigma concl)
@@ -481,7 +492,7 @@ let pr_goal ?(diffs=false) ?og_s g_s =
 
 (* display a goal tag *)
 let pr_goal_tag g =
-  let s = " (ID " ^ Goal.uid g ^ ")" in
+  let s = " (ID " ^ Proof.goal_uid g ^ ")" in
   str s
 
 (* display a goal name *)
@@ -494,12 +505,12 @@ let pr_goal_header nme sigma g =
   ++ (if should_gname() then str " " ++ Pp.surround (pr_existential_key (Global.env ()) sigma g) else mt ())
 
 (* display the conclusion of a goal *)
-let pr_concl n ?(diffs=false) ?og_s sigma g =
+let pr_concl n ?diffs sigma g =
   let env, concl = goal_repr sigma g in
-  let pc =
-    if diffs then
-      Proof_diffs.diff_concl ?og_s sigma g
-    else
+  let pc = match diffs with
+  | Some og_s ->
+      Proof_diffs.diff_concl ?og_s (Proof_diffs.make_goal env sigma g)
+  | None ->
       pr_letype_env ~goal_concl_style:true env sigma concl
   in
   let header = pr_goal_header (int n) sigma g in
@@ -569,7 +580,7 @@ let pr_ne_evar_set hd tl sigma l =
     mt ()
 
 let pr_selected_subgoal name sigma g =
-  let pg = pr_goal { sigma=sigma ; it=g; } in
+  let pg = pr_goal sigma g in
   let header = pr_goal_header name sigma g in
   v 0 (header ++ str " is:" ++ cut () ++ pg)
 
@@ -641,50 +652,63 @@ let should_print_dependent_evars =
     ~key:["Printing";"Dependent";"Evars";"Line"]
     ~value:false
 
+let gather_dependent_evars_goal sigma goals =
+  let map evk =
+    let evi = Evd.find sigma evk in
+    EConstr.mkEvar (evk, Evd.evar_identity_subst evi)
+  in
+  Evarutil.gather_dependent_evars sigma (List.map map goals)
+
+let print_dependent_evars_core gl sigma evars =
+  let mt_pp = mt () in
+  let evars_pp = Evar.Map.fold (fun e i s ->
+      let e' = pr_internal_existential_key e in
+      let sep = if s = mt_pp then "" else ", " in
+      s ++ str sep ++ e' ++
+      (match i with
+       | None -> str ":" ++ (Termops.pr_existential_key (Global.env ()) sigma e)
+       | Some i ->
+         let using = Evar.Set.fold (fun d s ->
+             s ++ str " " ++ (pr_internal_existential_key d))
+             i mt_pp in
+         str " using" ++ using))
+      evars mt_pp
+  in
+  let evars_current_pp = match gl with
+    | None -> mt_pp
+    | Some gl ->
+      let evars_current = gather_dependent_evars_goal sigma [gl] in
+      Evar.Map.fold (fun e _ s ->
+          s ++ str " " ++ (pr_internal_existential_key e))
+        evars_current mt_pp
+  in
+  cut () ++ cut () ++
+  str "(dependent evars: " ++ evars_pp ++
+  str "; in current goal:" ++ evars_current_pp ++ str ")"
+
+
 let print_dependent_evars gl sigma seeds =
   if should_print_dependent_evars () then
-    let mt_pp = mt () in
-    let evars = Evarutil.gather_dependent_evars sigma seeds in
-    let evars_pp = Evar.Map.fold (fun e i s ->
-        let e' = pr_internal_existential_key e in
-        let sep = if s = mt_pp then "" else ", " in
-        s ++ str sep ++ e' ++
-            (match i with
-            | None -> str ":" ++ (Termops.pr_existential_key (Global.env ()) sigma e)
-            | Some i ->
-              let using = Evar.Set.fold (fun d s ->
-                  s ++ str " " ++ (pr_internal_existential_key d))
-                i mt_pp in
-              str " using" ++ using))
-      evars mt_pp
-    in
-    let evars_current_pp = match gl with
-        | None -> mt_pp
-        | Some gl ->
-           let evars_current = Evarutil.gather_dependent_evars sigma [ gl ] in
-           Evar.Map.fold (fun e _ s ->
-               s ++ str " " ++ (pr_internal_existential_key e))
-             evars_current mt_pp
-    in
-    cut () ++ cut () ++
-      str "(dependent evars: " ++ evars_pp ++
-      str "; in current goal:" ++ evars_current_pp ++ str ")"
+    let evars = gather_dependent_evars_goal sigma seeds in
+    print_dependent_evars_core gl sigma evars
   else mt ()
 
-module GoalMap = Evar.Map
+let print_dependent_evars_entry gl sigma = function
+  | None -> mt ()
+  | Some entry ->
+    if should_print_dependent_evars () then
+      let terms = List.map pi2 (Proofview.initial_goals entry) in
+      let evars = Evarutil.gather_dependent_evars sigma terms in
+      print_dependent_evars_core gl sigma evars
+    else mt ()
 
 (* Print open subgoals. Checks for uninstantiated existential variables *)
-(* spiwack: [seeds] is for printing dependent evars in emacs mode. *)
+(* spiwack: [entry] is for printing dependent evars in emacs mode. *)
 (* spiwack: [pr_first] is true when the first goal must be singled out
    and printed in its entirety. *)
 (* [os_map] is derived from the previous proof step, used for diffs *)
-let pr_subgoals ?(pr_first=true) ?(diffs=false) ?os_map
-                        close_cmd sigma ~seeds ~shelf ~stack ~unfocused ~goals =
-  let diff_goal_map =
-    match os_map with
-    | Some (_, diff_goal_map) -> diff_goal_map
-    | None -> GoalMap.empty
-  in
+let pr_subgoals ?(pr_first=true) ?diffs ?entry
+    sigma ~shelf ~stack ~unfocused ~goals =
 
   (* Printing functions for the extra informations. *)
   let rec print_stack a = function
@@ -719,42 +743,30 @@ let pr_subgoals ?(pr_first=true) ?(diffs=false) ?os_map
     else str" " (* non-breakable space *)
   in
 
-  let get_ogs g =
-    match os_map with
-    | Some (osigma, _) ->
-      (* if Not_found, returning None treats the goal as new and it will be diff highlighted;
-         returning Some { it = g; sigma = sigma } will compare the new goal
-         to itself and it won't be highlighted *)
-      (try Some (GoalMap.find g diff_goal_map, osigma)
-      with Not_found -> None)
-    | None -> None
+  let get_ogs map g = match map with
+  | None -> None
+  | Some map -> Proof_diffs.map_goal g map
   in
   let rec pr_rec n = function
     | [] -> (mt ())
     | g::rest ->
-       let og_s = get_ogs g in
-       let pc = pr_concl n ~diffs ?og_s sigma g in
+      let diffs = Option.map (fun map -> get_ogs map g) diffs in
+      let pc = pr_concl n ?diffs sigma g in
         let prest = pr_rec (n+1) rest in
         (cut () ++ pc ++ prest)
   in
   let print_multiple_goals g l =
     if pr_first then
-      let og_s = get_ogs g in
-      pr_goal ~diffs ?og_s { it = g ; sigma = sigma }
+      let diffs = Option.map (fun map -> get_ogs map g) diffs in
+      pr_goal ?diffs sigma g
       ++ (if l=[] then mt () else cut ())
       ++ pr_rec 2 l
     else
       pr_rec 1 (g::l)
   in
-  let pr_evar_info gl sigma seeds =
+  let pr_evar_info gl =
     let first_goal = if pr_first then gl else None in
-    print_evar_constraints gl sigma ++ print_dependent_evars first_goal sigma seeds
-  in
-  (* Side effect! This has to be made more robust *)
-  let () =
-    match close_cmd with
-    | Some cmd -> Feedback.msg_info cmd
-    | None -> ()
+    print_evar_constraints gl sigma ++ print_dependent_evars_entry first_goal sigma entry
   in
 
   (* Main function *)
@@ -762,13 +774,13 @@ let pr_subgoals ?(pr_first=true) ?(diffs=false) ?os_map
   | [] ->
     let exl = Evd.undefined_map sigma in
     if Evar.Map.is_empty exl then
-      v 0 (str "No more goals." ++ pr_evar_info None sigma seeds)
+      v 0 (str "No more goals." ++ pr_evar_info None)
     else
       let pei = pr_evars_int sigma ~shelf ~given_up:[] 1 exl in
       v 0 ((str "No more goals,"
           ++ str " but there are non-instantiated existential variables:"
           ++ cut () ++ (hov 0 pei)
-          ++ pr_evar_info None sigma seeds
+          ++ pr_evar_info None
           ++ cut () ++ str "You can use Unshelve."))
   | g1::rest ->
       let goals = print_multiple_goals g1 rest in
@@ -782,65 +794,61 @@ let pr_subgoals ?(pr_first=true) ?(diffs=false) ?os_map
         ++ (if unfocused=[] then str ""
            else (cut() ++ cut() ++ str "*** Unfocused goals:" ++ cut()
                  ++ pr_rec (List.length rest + 2) unfocused))
-        ++ pr_evar_info (Some g1) sigma seeds
+        ++ pr_evar_info (Some g1)
       )
 
-let pr_open_subgoals_diff ?(quiet=false) ?(diffs=false) ?oproof proof =
+let pr_open_subgoals ?(quiet=false) ?diffs proof =
   (* spiwack: it shouldn't be the job of the printer to look up stuff
      in the [evar_map], I did stuff that way because it was more
      straightforward, but seriously, [Proof.proof] should return
      [evar_info]-s instead. *)
   let p = proof in
-  let Proof.{goals; stack; sigma} = Proof.data p in
+  let Proof.{goals; stack; sigma;entry} = Proof.data p in
   let shelf = Evd.shelf sigma in
   let given_up = Evd.given_up sigma in
   let stack = List.map (fun (l,r) -> List.length l + List.length r) stack in
-  let seeds = Proof.V82.top_evars p in
   begin match goals with
-  | [] -> let { Evd.it = bgoals ; sigma = bsigma } = Proof.V82.background_subgoals p in
+  | [] -> let bgoals = Proof.background_subgoals p in
           begin match bgoals,shelf,given_up with
-          | [] , [] , g when Evar.Set.is_empty g -> pr_subgoals None sigma ~seeds ~shelf ~stack ~unfocused:[] ~goals
+          | [] , [] , g when Evar.Set.is_empty g -> pr_subgoals sigma ~entry ~shelf ~stack ~unfocused:[] ~goals
           | [] , [] , _ ->
              Feedback.msg_info (str "No more goals, but there are some goals you gave up:");
              fnl ()
-            ++ pr_subgoals ~pr_first:false None bsigma ~seeds ~shelf:[] ~stack:[] ~unfocused:[] ~goals:(Evar.Set.elements given_up)
+            ++ pr_subgoals ~pr_first:false sigma ~entry ~shelf:[] ~stack:[] ~unfocused:[] ~goals:(Evar.Set.elements given_up)
             ++ fnl () ++ str "You need to go back and solve them."
           | [] , _ , _ ->
             Feedback.msg_info (str "All the remaining goals are on the shelf.");
             fnl ()
-            ++ pr_subgoals ~pr_first:false None bsigma ~seeds ~shelf:[] ~stack:[] ~unfocused:[] ~goals:shelf
+            ++ pr_subgoals ~pr_first:false sigma ~entry ~shelf:[] ~stack:[] ~unfocused:[] ~goals:shelf
           | _ , _, _ ->
-            let cmd = if quiet then None else
-              Some
+            let () =
+              if quiet then ()
+              else
+              Feedback.msg_info
                 (str "This subproof is complete, but there are some unfocused goals." ++
                 (let s = Proof_bullet.suggest p in
                  if Pp.ismt s then s else fnl () ++ s) ++
                 fnl ())
             in
-            pr_subgoals ~pr_first:false cmd bsigma ~seeds ~shelf ~stack:[] ~unfocused:[] ~goals:bgoals
+            pr_subgoals ~pr_first:false sigma ~entry ~shelf ~stack:[] ~unfocused:[] ~goals:bgoals
           end
   | _ ->
-     let { Evd.it = bgoals ; sigma = bsigma } = Proof.V82.background_subgoals p in
+     let bgoals = Proof.background_subgoals p in
      let bgoals_focused, bgoals_unfocused = List.partition (fun x -> List.mem x goals) bgoals in
      let unfocused_if_needed = if should_unfoc() then bgoals_unfocused else [] in
-     let os_map = match oproof with
-       | Some op when diffs ->
-         (try
-           let Proof.{sigma=osigma} = Proof.data op in
-           let diff_goal_map = Proof_diffs.make_goal_map oproof proof in
-           Some (osigma, diff_goal_map)
+     let diffs = match diffs with
+       | Some (Some op) ->
+         Some (try Some (Proof_diffs.make_goal_map op proof)
          with Pp_diff.Diff_Failure msg ->
            Proof_diffs.notify_proof_diff_failure msg;
            None)
-       | _ -> None
+       | Some None -> Some None
+       | None -> None
      in
-     pr_subgoals ~pr_first:true ~diffs ?os_map None bsigma ~seeds ~shelf ~stack:[]
+     pr_subgoals ~pr_first:true ?diffs sigma ~entry ~shelf ~stack:[]
         ~unfocused:unfocused_if_needed ~goals:bgoals_focused
   end
 
-let pr_open_subgoals ~proof =
-  pr_open_subgoals_diff proof
-
 let pr_nth_open_subgoal ~proof n =
   let Proof.{goals;sigma} = Proof.data proof in
   pr_subgoal n sigma goals
@@ -860,13 +868,13 @@ let pr_goal_emacs ~proof gid sid =
   match proof with
   | None -> user_err Pp.(str "No proof for that state.")
   | Some proof ->
-    let pr gs =
+    let pr sigma gs =
       v 0 ((str "goal ID " ++ (int gid) ++ str " at state " ++ (int sid)) ++ cut ()
-          ++ pr_goal gs)
+          ++ pr_goal sigma gs)
     in
     try
       let { Proof.sigma } = Proof.data proof in
-      pr { it = Evar.unsafe_of_int gid ; sigma }
+      pr sigma (Evar.unsafe_of_int gid)
     with Not_found -> user_err Pp.(str "No such goal.")
 
 (* Printer function for sets of Assumptions.assumptions.
@@ -1026,28 +1034,17 @@ let pr_assumptionset env sigma s =
     ] in
     prlist_with_sep fnl (fun x -> x) (Option.List.flatten assums)
 
-(* print the proof step, possibly with diffs highlighted, *)
-let print_and_diff oldp newp =
-  match newp with
-  | None -> ()
-  | Some proof ->
-    let output =
-      if Proof_diffs.show_diffs () then
-        try pr_open_subgoals_diff ~diffs:true ?oproof:oldp proof
-        with Pp_diff.Diff_Failure msg -> begin
-          (* todo: print the unparsable string (if we know it) *)
-          Feedback.msg_warning Pp.(str ("Diff failure: " ^ msg) ++ cut()
-              ++ str "Showing results without diff highlighting" );
-          pr_open_subgoals ~proof
-        end
-      else
-        pr_open_subgoals ~proof
-    in
-    Feedback.msg_notice output
-
 let pr_typing_flags flags =
   str "check_guarded: " ++ bool flags.check_guarded ++ fnl ()
   ++ str "check_positive: " ++ bool flags.check_positive ++ fnl ()
   ++ str "check_universes: " ++ bool flags.check_universes ++ fnl ()
   ++ str "cumulative sprop: " ++ bool flags.cumulative_sprop ++ fnl ()
   ++ str "definitional uip: " ++ bool flags.allow_uip
+
+module Debug =
+struct
+
+let pr_goal gl =
+  pr_goal (Proofview.Goal.sigma gl) (Proofview.Goal.goal gl)
+
+end
