@@ -13,7 +13,6 @@ open Constr
 open EConstr
 open Environ
 open Evd
-open Clenv
 open Redexpr
 open Pattern
 open Unification
@@ -113,6 +112,9 @@ val force_destruction_arg : evars_flag -> env -> evar_map ->
     delayed_open_constr_with_bindings destruction_arg ->
     evar_map * constr with_bindings destruction_arg
 
+val finish_evar_resolution : ?flags:Pretyping.inference_flags ->
+  env -> evar_map -> (evar_map * constr) -> evar_map * constr
+
 (** Tell if a used hypothesis should be cleared by default or not *)
 
 val use_clear_hyp_by_default : unit -> bool
@@ -281,7 +283,7 @@ val general_elim  : evars_flag -> clear_flag ->
   constr with_bindings -> eliminator -> unit Proofview.tactic
 
 val general_elim_clause : evars_flag -> unify_flags -> Id.t option ->
-  clausenv -> eliminator -> unit Proofview.tactic
+  (EConstr.t * EConstr.t) -> eliminator -> unit Proofview.tactic
 
 val default_elim  : evars_flag -> clear_flag -> constr with_bindings ->
   unit Proofview.tactic
