@@ -1,21 +1,13 @@
 (************************************************************************)
-(*  v      *   The Coq Proof Assistant  /  The Coq Development Team     *)
-(* <O___,, *   INRIA - CNRS - LIX - LRI - PPS - Copyright 1999-2016     *)
+(*         *   The Coq Proof Assistant / The Coq Development Team       *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
-(*    //   *      This file is distributed under the terms of the       *)
-(*         *       GNU Lesser General Public License Version 2.1        *)
+(*    //   *    This file is distributed under the terms of the         *)
+(*         *     GNU Lesser General Public License Version 2.1          *)
+(*         *     (see LICENSE file for the text of the license)         *)
 (************************************************************************)
 
-(* Modified by Lionel Elie Mamane <lionel@mamane.lu> on 9 & 10 Mar 2004:
- *  - handling of absolute filenames (function coq_module)
- *  - coq_module: chop ./// (arbitrary amount of slashes), not only "./"
- *  - function chop_prefix not useful anymore. Deleted.
- *  - correct typo in usage message: "-R" -> "--R"
- *  - shorten the definition of make_path
- * This notice is made to comply with section 2.a of the GPLv2.
- * It may be removed or abbreviated as far as I am concerned.
- *)
-
 open Cdglobals
 open Printf
 
@@ -55,10 +47,9 @@ let usage () =
   prerr_endline "  --verbose            verbose mode";
   prerr_endline "  --no-externals       no links to Coq standard library";
   prerr_endline "  --external <url> <d> set URL for external library d";
-  prerr_endline "  --coqlib <url>       set URL for Coq standard library";
+  prerr_endline "  --coqlib_url <url>   set URL for Coq standard library";
   prerr_endline ("                       (default is " ^ Coq_config.wwwstdlib ^ ")");
-  prerr_endline "  --boot               run in boot mode";
-  prerr_endline "  --coqlib_path <dir>  set the path where Coq files are installed";
+  prerr_endline "  --coqlib <dir>       set the path where Coq files are installed";
   prerr_endline "  -R <dir> <coqdir>    map physical dir to Coq dir";
   prerr_endline "  -Q <dir> <coqdir>    map physical dir to Coq dir";
   prerr_endline "  --latin1             set ISO-8859-1 mode";
@@ -76,17 +67,13 @@ let usage () =
   prerr_endline "";
   exit 1
 
-let obsolete s =
-  eprintf "Warning: option %s is now obsolete; please update your scripts\n" s
-
 (*s \textbf{Banner.} Always printed. Notice that it is printed on error
     output, so that when the output of [coqdoc] is redirected this header
     is not (unless both standard and error outputs are redirected, of
     course). *)
 
 let banner () =
-  eprintf "This is coqdoc version %s, compiled on %s\n"
-    Coq_config.version Coq_config.compile_date;
+  eprintf "This is coqdoc version %s\n" Coq_config.version;
   flush stderr
 
 let target_full_name f =
@@ -100,13 +87,6 @@ let target_full_name f =
     suffix.  Coq files have suffixe \verb!.v! or \verb!.g! and \LaTeX\
     files have suffix \verb!.tex!. *)
 
-let check_if_file_exists f =
-  if not (Sys.file_exists f) then begin
-    eprintf "coqdoc: %s: no such file\n" f;
-    exit 1
-  end
-
-
 (* [paths] maps a physical path to a name *)
 let paths = ref []
 
@@ -125,18 +105,21 @@ let rec name_of_path p name dirname suffix =
 let coq_module filename =
   let bfname = Filename.chop_extension filename in
   let dirname, fname = normalize_filename bfname in
+  let _ = match Unicode.ident_refutation fname with
+    | Some err -> eprintf "\ncoqdoc: not a valid filename %s.v\n" fname; exit 1
+    | None -> () in
   let rec change_prefix = function
     (* Follow coqc: if in scope of -R, substitute logical name *)
     (* otherwise, keep only base name *)
     | [] -> fname
     | (p, name) :: rem ->
-	try name_of_path p name dirname [fname]
-	with Not_found -> change_prefix rem
+        try name_of_path p name dirname [fname]
+        with Not_found -> change_prefix rem
   in
   change_prefix !paths
 
 let what_file f =
-  check_if_file_exists f;
+  FileUtil.check_if_file_exists f;
   if Filename.check_suffix f ".v" || Filename.check_suffix f ".g" then
     Vernac_file (f, coq_module f)
   else if Filename.check_suffix f ".tex" then
@@ -144,45 +127,22 @@ let what_file f =
   else
      (eprintf "\ncoqdoc: don't know what to do with %s\n" f; exit 1)
 
-(*s \textbf{Reading file names from a file.}
- *  File names may be given
- *  in a file instead of being given on the command
- *  line. [(files_from_file f)] returns the list of file names contained
- *  in the file named [f]. These file names must be separated by spaces,
- *  tabulations or newlines.
- *)
-
-let files_from_file f =
-  let files_from_channel ch =
-    let buf = Buffer.create 80 in
-    let l = ref [] in
-      try
-	while true do
-	  match input_char ch with
-	    | ' ' | '\t' | '\n' ->
-		if Buffer.length buf > 0 then l := (Buffer.contents buf) :: !l;
-		Buffer.clear buf
-	    | c ->
-		Buffer.add_char buf c
-	done; []
-      with End_of_file ->
-	List.rev !l
-  in
-    try
-      check_if_file_exists f;
-      let ch = open_in f in
-      let l = files_from_channel ch in
-	close_in ch;l
-    with Sys_error s -> begin
-      eprintf "coqdoc: cannot read from file %s (%s)\n" f s;
-      exit 1
-    end
-
 (*s \textbf{Parsing of the command line.} *)
 
-let dvi = ref false
-let ps = ref false
-let pdf = ref false
+let compile_targets = ref []
+type glob_source_t = NoGlob | DotGlob | GlobFile of string
+let glob_source = ref DotGlob
+
+let warn_on_option_renaming_for_url d =
+  try
+    let https_prefix = String.sub d 0 4 in
+    if String.equal https_prefix "http"
+    then
+      Format.eprintf
+        "warning: use --coqlib_url to specify stdlib's URL starting with Coq 8.14@\n%!"
+    else ()
+  with Invalid_argument _ ->
+    ()
 
 let parse () =
   let files = ref [] in
@@ -192,79 +152,86 @@ let parse () =
 
     | ("-nopreamble" | "--nopreamble" | "--no-preamble"
       |  "-bodyonly"   | "--bodyonly"   | "--body-only") :: rem ->
-	header_trailer := false; parse_rec rem
+        header_trailer := false; parse_rec rem
     | ("-with-header" | "--with-header") :: f ::rem ->
-	header_trailer := true; header_file_spec := true; header_file := f; parse_rec rem
+        header_trailer := true; header_file_spec := true; header_file := f; parse_rec rem
     | ("-with-header" | "--with-header") :: [] ->
-	usage ()
+        usage ()
     | ("-with-footer" | "--with-footer") :: f ::rem ->
-	header_trailer := true; footer_file_spec := true; footer_file := f; parse_rec rem
+        header_trailer := true; footer_file_spec := true; footer_file := f; parse_rec rem
     | ("-with-footer" | "--with-footer") :: [] ->
-	usage ()
+        usage ()
     | ("-p" | "--preamble") :: s :: rem ->
-	Output.push_in_preamble s; parse_rec rem
+        Output.push_in_preamble s; parse_rec rem
     | ("-p" | "--preamble") :: [] ->
-	usage ()
+        usage ()
     | ("-noindex" | "--noindex" | "--no-index") :: rem ->
-	index := false; parse_rec rem
+        index := false; parse_rec rem
     | ("-multi-index" | "--multi-index") :: rem ->
-	multi_index := true; parse_rec rem
+        multi_index := true; parse_rec rem
     | ("-index" | "--index") :: s :: rem ->
-	Cdglobals.index_name := s; parse_rec rem
+        Cdglobals.index_name := s; parse_rec rem
     | ("-index" | "--index") :: [] ->
-	usage ()
+        usage ()
     | ("-toc" | "--toc" | "--table-of-contents") :: rem ->
-	toc := true; parse_rec rem
+        toc := true; parse_rec rem
     | ("-stdout" | "--stdout") :: rem ->
-	out_to := StdOut; parse_rec rem
+        out_to := StdOut; parse_rec rem
     | ("-o" | "--output") :: f :: rem ->
-	out_to := File (Filename.basename f); output_dir := Filename.dirname f; parse_rec rem
+        out_to := File (Filename.basename f); output_dir := Filename.dirname f; parse_rec rem
     | ("-o" | "--output") :: [] ->
-	usage ()
+        usage ()
     | ("-d" | "--directory") :: dir :: rem ->
-	output_dir := dir; parse_rec rem
+        output_dir := dir; parse_rec rem
     | ("-d" | "--directory") :: [] ->
-	usage ()
+        usage ()
     | ("-s" | "--short") :: rem ->
-	short := true; parse_rec rem
+        short := true; parse_rec rem
     | ("-l" | "-light" | "--light") :: rem ->
-	gallina := true; light := true; parse_rec rem
+        gallina := true; light := true; parse_rec rem
     | ("-g" | "-gallina" | "--gallina") :: rem ->
-	gallina := true; parse_rec rem
+        gallina := true; parse_rec rem
     | ("-t" | "-title" | "--title") :: s :: rem ->
-	title := s; parse_rec rem
+        title := s; parse_rec rem
     | ("-t" | "-title" | "--title") :: [] ->
-	usage ()
+        usage ()
     | ("-latex" | "--latex") :: rem ->
-	Cdglobals.target_language := LaTeX; parse_rec rem
+        Cdglobals.target_language := LaTeX;
+        parse_rec rem
     | ("-pdf" | "--pdf") :: rem ->
-	Cdglobals.target_language := LaTeX; pdf := true; parse_rec rem
+        Cdglobals.target_language := LaTeX;
+        compile_targets := LatexCompiler.Pdf :: !compile_targets;
+        parse_rec rem
     | ("-dvi" | "--dvi") :: rem ->
-	Cdglobals.target_language := LaTeX; dvi := true; parse_rec rem
+        Cdglobals.target_language := LaTeX;
+        compile_targets := LatexCompiler.Dvi :: !compile_targets;
+        parse_rec rem
     | ("-ps" | "--ps") :: rem ->
-	Cdglobals.target_language := LaTeX; ps := true; parse_rec rem
+        Cdglobals.target_language := LaTeX;
+        compile_targets := LatexCompiler.Ps :: !compile_targets;
+        parse_rec rem
     | ("-html" | "--html") :: rem ->
-	Cdglobals.target_language := HTML; parse_rec rem
+        Cdglobals.target_language := HTML; parse_rec rem
     | ("-texmacs" | "--texmacs") :: rem ->
-	Cdglobals.target_language := TeXmacs; parse_rec rem
+        Cdglobals.target_language := TeXmacs; parse_rec rem
     | ("-raw" | "--raw") :: rem ->
-	Cdglobals.target_language := Raw; parse_rec rem
+        Cdglobals.target_language := Raw; parse_rec rem
     | ("-charset" | "--charset") :: s :: rem ->
-	Cdglobals.charset := s; parse_rec rem
+        Cdglobals.charset := s; parse_rec rem
     | ("-charset" | "--charset") :: [] ->
-	usage ()
+        usage ()
     | ("-inputenc" | "--inputenc") :: s :: rem ->
-	Cdglobals.inputenc := s; parse_rec rem
+        Cdglobals.inputenc := s; parse_rec rem
     | ("-inputenc" | "--inputenc") :: [] ->
-	usage ()
+        usage ()
     | ("-raw-comments" | "--raw-comments") :: rem ->
-	Cdglobals.raw_comments := true; parse_rec rem
+        Cdglobals.raw_comments := true; parse_rec rem
     | ("-parse-comments" | "--parse-comments") :: rem ->
-	Cdglobals.parse_comments := true; parse_rec rem
+        Cdglobals.parse_comments := true; parse_rec rem
     | ("-plain-comments" | "--plain-comments") :: rem ->
-	Cdglobals.plain_comments := true; parse_rec rem
+        Cdglobals.plain_comments := true; parse_rec rem
     | ("-interpolate" | "--interpolate") :: rem ->
-	Cdglobals.interpolate := true; parse_rec rem
+        Cdglobals.interpolate := true; parse_rec rem
     | ("-toc-depth" | "--toc-depth") :: [] ->
       usage ()
     | ("-toc-depth" | "--toc-depth") :: ds :: rem ->
@@ -289,68 +256,64 @@ let parse () =
       parse_rec rem
 
     | ("-latin1" | "--latin1") :: rem ->
-	Cdglobals.set_latin1 (); parse_rec rem
+        Cdglobals.set_latin1 (); parse_rec rem
     | ("-utf8" | "--utf8") :: rem ->
-	Cdglobals.set_utf8 (); parse_rec rem
+        Cdglobals.set_utf8 (); parse_rec rem
 
     | ("-q" | "-quiet" | "--quiet") :: rem ->
-	quiet := true; parse_rec rem
+        quiet := true; parse_rec rem
     | ("-v" | "-verbose" | "--verbose") :: rem ->
-	quiet := false; parse_rec rem
+        quiet := false; parse_rec rem
 
     | ("-h" | "-help" | "-?" | "--help") :: rem ->
-	banner (); usage ()
+        banner (); usage ()
     | ("-V" | "-version" | "--version") :: _ ->
-	banner (); exit 0
+        banner (); exit 0
 
     | ("-vernac-file" | "--vernac-file") :: f :: rem ->
-	check_if_file_exists f;
-	add_file (Vernac_file (f, coq_module f)); parse_rec rem
+        FileUtil.check_if_file_exists f;
+        add_file (Vernac_file (f, coq_module f)); parse_rec rem
     | ("-vernac-file" | "--vernac-file") :: [] ->
-	usage ()
+        usage ()
     | ("-tex-file" | "--tex-file") :: f :: rem ->
-	add_file (Latex_file f); parse_rec rem
+        add_file (Latex_file f); parse_rec rem
     | ("-tex-file" | "--tex-file") :: [] ->
-	usage ()
+        usage ()
     | ("-files" | "--files" | "--files-from") :: f :: rem ->
-	List.iter (fun f -> add_file (what_file f)) (files_from_file f);
-	parse_rec rem
+        List.iter (fun f -> add_file (what_file f)) (FileUtil.files_from_file f);
+        parse_rec rem
     | ("-files" | "--files") :: [] ->
-	usage ()
+        usage ()
     | "-R" :: path :: log :: rem ->
-	add_path path log; parse_rec rem
+        add_path path log; parse_rec rem
     | "-R" :: ([] | [_]) ->
-	usage ()
+        usage ()
     | "-Q" :: path :: log :: rem ->
-	add_path path log; parse_rec rem
+        add_path path log; parse_rec rem
     | "-Q" :: ([] | [_]) ->
-	usage ()
+        usage ()
     | ("-glob-from" | "--glob-from") :: f :: rem ->
-	glob_source := GlobFile f; parse_rec rem
+        glob_source := GlobFile f; parse_rec rem
     | ("-glob-from" | "--glob-from") :: [] ->
-	usage ()
+        usage ()
     | ("-no-glob" | "--no-glob") :: rem ->
-	glob_source := NoGlob; parse_rec rem
+        glob_source := NoGlob; parse_rec rem
     | ("--no-externals" | "-no-externals" | "-noexternals") :: rem ->
-	Cdglobals.externals := false; parse_rec rem
+        Cdglobals.externals := false; parse_rec rem
     | ("--external" | "-external") :: u :: logicalpath :: rem ->
-	Index.add_external_library logicalpath u; parse_rec rem
-    | ("--coqlib" | "-coqlib") :: u :: rem ->
-	Cdglobals.coqlib := u; parse_rec rem
+        Index.add_external_library logicalpath u; parse_rec rem
+    | ("--coqlib_url" | "-coqlib_url") :: u :: rem ->
+        Cdglobals.coqlib_url := u; parse_rec rem
+    | ("--coqlib_url" | "-coqlib_url") :: [] ->
+        usage ()
+    | ("--coqlib" | "-coqlib") :: d :: rem ->
+        warn_on_option_renaming_for_url d;
+        Boot.Env.set_coqlib d; parse_rec rem
     | ("--coqlib" | "-coqlib") :: [] ->
-	usage ()
-    | ("--boot" | "-boot") :: rem ->
-	Cdglobals.coqlib_path := normalize_path (
-          Filename.concat
-            (Filename.dirname Sys.executable_name)
-            Filename.parent_dir_name
-        ); parse_rec rem
-    | ("--coqlib_path" | "-coqlib_path") :: d :: rem ->
-	Cdglobals.coqlib_path := d; parse_rec rem
-    | ("--coqlib_path" | "-coqlib_path") :: [] ->
-	usage ()
+        usage ()
     | f :: rem ->
-	add_file (what_file f); parse_rec rem
+      if String.length f > 0 && f.[0] = '-' then usage()
+      else add_file (what_file f); parse_rec rem
   in
     parse_rec (List.tl (Array.to_list Sys.argv));
     List.rev !files
@@ -361,55 +324,14 @@ let parse () =
     If option \verb!-dvi!, \verb!-ps! or \verb!-html! is invoked, then
     we make calls to \verb!latex! or \verb!dvips! or \verb!pdflatex! accordingly. *)
 
-let locally dir f x =
-  let cwd = Sys.getcwd () in
-    try
-      Sys.chdir dir; let y = f x in Sys.chdir cwd; y
-    with e ->
-      Sys.chdir cwd; raise e
-
-let clean_temp_files basefile =
-  let remove f = try Sys.remove f with _ -> () in
-    remove (basefile ^ ".tex");
-    remove (basefile ^ ".log");
-    remove (basefile ^ ".aux");
-    remove (basefile ^ ".toc");
-    remove (basefile ^ ".dvi");
-    remove (basefile ^ ".ps");
-    remove (basefile ^ ".pdf");
-    remove (basefile ^ ".haux");
-    remove (basefile ^ ".html")
-
-let clean_and_exit file res = clean_temp_files file; exit res
-
-let cat file =
-  let c = open_in file in
-    try
-      while true do print_char (input_char c) done
-    with End_of_file ->
-      close_in c
-
-let copy src dst =
-  let cin = open_in src in
-  try
-    let cout = open_out dst in
-    try
-      while true do Pervasives.output_char cout (input_char cin) done
-    with End_of_file ->
-      close_out cout;
-      close_in cin
-  with Sys_error e ->
-    eprintf "%s\n" e;
-    exit 1
-
 (*s Functions for generating output files *)
 
 let gen_one_file l =
   let file = function
     | Vernac_file (f,m) ->
-        let sub = if !lib_subtitles then Cpretty.detect_subtitle f m else None in
-          Output.set_module m sub;
-          Cpretty.coq_file f m
+      let sub = if !lib_subtitles then Cpretty.detect_subtitle f m else None in
+      Output.set_module m sub;
+      Cpretty.coq_file f m
     | Latex_file _ -> ()
   in
     if (!header_trailer) then Output.header ();
@@ -422,13 +344,13 @@ let gen_mult_files l =
   let file = function
     | Vernac_file (f,m) ->
       let sub = if !lib_subtitles then Cpretty.detect_subtitle f m else None in
-	let hf = target_full_name m in
+        let hf = target_full_name m in
         Output.set_module m sub;
-	  open_out_file hf;
-	  if (!header_trailer) then Output.header ();
-	  Cpretty.coq_file f m;
-	  if (!header_trailer) then Output.trailer ();
-	  close_out_file()
+          open_out_file hf;
+          if (!header_trailer) then Output.header ();
+          Cpretty.coq_file f m;
+          if (!header_trailer) then Output.trailer ();
+          close_out_file()
     | Latex_file _ -> ()
   in
     List.iter file l;
@@ -453,112 +375,67 @@ let gen_mult_files l =
     (* NB: for latex and texmacs, a separated toc or index is meaningless... *)
 
 let read_glob_file vfile f =
-  try Index.read_glob vfile f
+  try Glob_file.read_glob vfile f
   with Sys_error s -> eprintf "Warning: %s (links will not be available)\n" s
 
 let read_glob_file_of = function
   | Vernac_file (f,_) ->
-      read_glob_file (Some f) (Filename.chop_extension f ^ ".glob")
+    read_glob_file (Some f) (Filename.chop_extension f ^ ".glob")
   | Latex_file _ -> ()
 
 let index_module = function
   | Vernac_file (f,m) ->
-      Index.add_module m
+    Index.add_module m
   | Latex_file _ -> ()
 
+module E = Boot.Env
+
 let copy_style_file file =
-  let src =
-    List.fold_left
-      Filename.concat !Cdglobals.coqlib_path ["tools";"coqdoc";file] in
+  (* We give preference to coqlib in case it is overriden *)
+  let env = E.init () in
+  let coqdoc = E.tool env "coqdoc" in
+  let sty_file = E.Path.relative coqdoc file in
+  if not (E.Path.exists sty_file) then
+    begin
+      let sty_file = E.Path.to_string sty_file in
+      eprintf "coqdoc: cannot find coqdoc style file: %s\n" sty_file;
+      exit 1
+    end;
+  let sty_file_s = E.Path.to_string sty_file in
   let dst = coqdoc_out file in
-  if Sys.file_exists src then copy src dst
-  else eprintf "Warning: file %s does not exist\n" src
+  FileUtil.copy sty_file_s dst
 
 let produce_document l =
   if !target_language=HTML then copy_style_file "coqdoc.css";
   if !target_language=LaTeX then copy_style_file "coqdoc.sty";
-  (match !Cdglobals.glob_source with
+  (match !glob_source with
     | NoGlob -> ()
     | DotGlob -> List.iter read_glob_file_of l
     | GlobFile f -> read_glob_file None f);
   List.iter index_module l;
   match !out_to with
     | StdOut ->
-	Cdglobals.out_channel := stdout;
-	gen_one_file l
+        Cdglobals.out_channel := stdout;
+        gen_one_file l
     | File f ->
-	open_out_file f;
-	gen_one_file l;
-	close_out_file()
+        open_out_file f;
+        gen_one_file l;
+        close_out_file()
     | MultFiles ->
-	gen_mult_files l
+        gen_mult_files l
 
 let produce_output fl =
-  if not (!dvi || !ps || !pdf) then
+  if List.length !compile_targets = 0 then
     produce_document fl
   else
-    begin
-      let texfile = Filename.temp_file "coqdoc" ".tex" in
-      let basefile = Filename.chop_suffix texfile ".tex" in
-      let final_out_to = !out_to in
-	out_to := File texfile;
-	output_dir := (Filename.dirname texfile);
-	produce_document fl;
-
-	let latexexe = if !pdf then "pdflatex" else "latex" in
-	let latexcmd =
-	  let file = Filename.basename texfile in
-	  let file =
-	    if !quiet then sprintf "'\\nonstopmode\\input{%s}'" file else file
-	  in
-	    sprintf "%s %s && %s %s 1>&2 %s" latexexe file latexexe file (if !quiet then "> /dev/null" else "")
-	in
-	let res = locally (Filename.dirname texfile) Sys.command latexcmd in
-	  if res <> 0 then begin
-	      eprintf "Couldn't run LaTeX successfully\n";
-	      clean_and_exit basefile res
-	    end;
-
-	  let dvifile = basefile ^ ".dvi" in
-	    if !dvi then
-	      begin
-		match final_out_to with
-		  | MultFiles | StdOut -> cat dvifile
-		  | File f -> copy dvifile f
-	      end;
-	  let pdffile = basefile ^ ".pdf" in
-	    if !pdf then
-            begin
-	        match final_out_to with
-		  | MultFiles | StdOut -> cat pdffile
-		  | File f -> copy pdffile f
-	    end;
-	    if !ps then begin
-	        let psfile = basefile ^ ".ps"
-		in
-		let command =
-		  sprintf "dvips %s -o %s %s" dvifile psfile
-		    (if !quiet then "> /dev/null 2>&1" else "")
-		in
-		let res = Sys.command command in
-		  if res <> 0 then begin
-		      eprintf "Couldn't run dvips successfully\n";
-		      clean_and_exit basefile res
-		    end;
-		  match final_out_to with
-		    | MultFiles | StdOut -> cat psfile
-		    | File f -> copy psfile f
-	      end;
-
-	    clean_temp_files basefile
-    end
-
+    let otypes = !compile_targets in
+    LatexCompiler.compile ~otypes ~produce_document fl
 
 (*s \textbf{Main program.} Print the banner, parse the command line,
     read the files and then call [produce_document] from module [Web]. *)
 
 let _ =
   let files = parse () in
-    Index.init_coqlib_library ();
+  Index.init_coqlib_library ();
     if not !quiet then banner ();
     if files <> [] then produce_output files
