@@ -9,7 +9,6 @@
 (************************************************************************)
 
 (*i*)
-module CVars = Vars
 open Names
 open EConstr
 open CErrors
@@ -25,7 +24,6 @@ open Class_tactics
 open Libobject
 
 module RelDecl = Context.Rel.Declaration
-module NamedDecl = Context.Named.Declaration
 (*i*)
 
 let set_typeclass_transparency ~locality c b =
@@ -42,7 +40,8 @@ let warn_deprecated_tc_transparency_without_locality =
    sections without specifying an explicit locality attribute is \
    therefore deprecated. It is recommended to use \"export\" whenever \
    possible. Use the attributes #[local], #[global] and #[export] \
-   depending on your choice. For example: \"#[export] Typeclasses Transparent foo.\"")
+   depending on your choice. For example: \"#[export] Typeclasses Transparent foo.\". \
+   This is supported since Coq 8.15.")
 
 let default_tc_transparency_locality () =
   if Global.sections_are_opened () then Hints.Local
@@ -103,14 +102,14 @@ let perform_instance i =
   let i = { is_class = i.inst_class; is_info = i.inst_info; is_impl = i.inst_impl } in
   Typeclasses.load_instance i
 
-let cache_instance (_, inst) = perform_instance inst
+let cache_instance inst = perform_instance inst
 
-let load_instance _ (_, inst) = match inst.inst_global with
+let load_instance _ inst = match inst.inst_global with
 | Local -> assert false
 | SuperGlobal -> perform_instance inst
 | Export -> ()
 
-let open_instance i (_, inst) = match inst.inst_global with
+let open_instance i inst = match inst.inst_global with
 | Local -> assert false
 | SuperGlobal -> perform_instance inst
 | Export -> if Int.equal i 1 then perform_instance inst
@@ -120,7 +119,7 @@ let subst_instance (subst, inst) =
       inst_class = fst (subst_global subst inst.inst_class);
       inst_impl = fst (subst_global subst inst.inst_impl) }
 
-let discharge_instance (_, inst) =
+let discharge_instance inst =
   match inst.inst_global with
   | Local -> None
   | SuperGlobal | Export ->
@@ -133,7 +132,7 @@ let rebuild_instance inst =
 
 let classify_instance inst = match inst.inst_global with
 | Local -> Dispose
-| SuperGlobal | Export -> Substitute inst
+| SuperGlobal | Export -> Substitute
 
 let instance_input : instance_obj -> obj =
   declare_object
@@ -182,7 +181,7 @@ let add_instance cl info global impl =
     inst_global = global ;
     inst_impl = impl;
   } in
-  Lib.add_anonymous_leaf (instance_input i);
+  Lib.add_leaf (instance_input i);
   add_instance_base i
 
 let warning_not_a_class =
@@ -210,7 +209,7 @@ let declare_instance ?(warn = false) env sigma info local glob =
  * classes persistent object
  *)
 
-let cache_class (_,c) = load_class c
+let cache_class c = load_class c
 
 let subst_class (subst,cl) =
   let do_subst_con c = Mod_subst.subst_constant subst c
@@ -236,39 +235,12 @@ let subst_class (subst,cl) =
     cl_strict = cl.cl_strict;
     cl_unique = cl.cl_unique }
 
-let discharge_class (_,cl) =
-  let open CVars in
-  let repl = Lib.replacement_context () in
-  let rel_of_variable_context ctx = List.fold_right
-    ( fun decl (ctx', subst) ->
-        let decl' = decl |> NamedDecl.map_constr (substn_vars 1 subst) |> NamedDecl.to_rel_decl in
-        (decl' :: ctx', NamedDecl.get_id decl :: subst)
-    ) ctx ([], []) in
-  let discharge_rel_context (subst, usubst) n rel =
-    let rel = Context.Rel.map (Cooking.expmod_constr repl) rel in
-    let fold decl (ctx, k) =
-      let map c = subst_univs_level_constr usubst (substn_vars k subst c) in
-      RelDecl.map_constr map decl :: ctx, succ k
-    in
-    let ctx, _ = List.fold_right fold rel ([], n) in
-    ctx
-  in
-  let abs_context cl =
-    let open GlobRef in
-    match cl.cl_impl with
-      | VarRef _ | ConstructRef _ -> assert false
-      | ConstRef cst -> Lib.section_segment_of_constant cst
-      | IndRef (ind,_) -> Lib.section_segment_of_mutual_inductive ind in
-  let discharge_context ctx' subst ctx =
-    discharge_rel_context subst 1 ctx @ ctx'
-  in
+let discharge_class cl =
   try
-    let info = abs_context cl in
-    let ctx = info.Declarations.abstr_ctx in
-    let ctx, subst = rel_of_variable_context ctx in
-    let usubst, cl_univs' = Lib.discharge_abstract_universe_context info cl.cl_univs in
-    let context = discharge_context ctx (subst, usubst) cl.cl_context in
-    let props = discharge_rel_context (subst, usubst) (succ (List.length cl.cl_context)) cl.cl_props in
+    let info = Lib.section_segment_of_reference cl.cl_impl in
+    let info, _, cl_univs' = Cooking.lift_poly_univs info cl.cl_univs in
+    let nprops = List.length cl.cl_props in
+    let props, context = List.chop nprops (Discharge.cook_rel_context info (cl.cl_props @ cl.cl_context)) in
     let discharge_proj x = x in
     { cl_univs = cl_univs';
       cl_impl = cl.cl_impl;
@@ -292,13 +264,13 @@ let class_input : typeclass -> obj =
     { (default_object "type classes state") with
       cache_function = cache_class;
       load_function = (fun _ -> cache_class);
-      classify_function = (fun x -> Substitute x);
+      classify_function = (fun x -> Substitute);
       discharge_function = (fun a -> Some (discharge_class a));
       rebuild_function = rebuild_class;
       subst_function = subst_class }
 
 let add_class cl =
-  Lib.add_anonymous_leaf (class_input cl)
+  Lib.add_leaf (class_input cl)
 
 let add_class env sigma cl =
   add_class cl;
@@ -366,9 +338,8 @@ let instance_hook info global ?hook cst =
 
 let declare_instance_constant iinfo global impargs ?hook name udecl poly sigma term termtype =
   let kind = Decls.(IsDefinition Instance) in
-  let scope = Locality.Global Locality.ImportDefaultBehavior in
   let cinfo = Declare.CInfo.make ~name ~impargs ~typ:(Some termtype) () in
-  let info = Declare.Info.make ~kind ~scope ~poly ~udecl () in
+  let info = Declare.Info.make ~kind ~poly ~udecl () in
   let kn = Declare.declare_definition ~cinfo ~info ~opaque:false ~body:term sigma in
   instance_hook iinfo global ?hook kn
 
@@ -397,10 +368,9 @@ let declare_instance_program pm env sigma ~locality ~poly name pri impargs udecl
   let obls, _, term, typ = RetrieveObl.retrieve_obligations env name sigma 0 term termtype in
   let hook = Declare.Hook.make hook in
   let uctx = Evd.evar_universe_context sigma in
-  let scope, kind = Locality.Global Locality.ImportDefaultBehavior,
-                    Decls.IsDefinition Decls.Instance in
+  let kind = Decls.IsDefinition Decls.Instance in
   let cinfo = Declare.CInfo.make ~name ~typ ~impargs () in
-  let info = Declare.Info.make  ~udecl ~scope ~poly ~kind ~hook () in
+  let info = Declare.Info.make  ~udecl ~poly ~kind ~hook () in
   let pm, _ =
     Declare.Obls.add_definition ~pm ~cinfo ~info ~term ~uctx obls
   in pm
