@@ -35,11 +35,11 @@ type object_pr = {
   print_inductive           : MutInd.t -> UnivNames.univ_name_list option -> Pp.t;
   print_constant_with_infos : Constant.t -> UnivNames.univ_name_list option -> Pp.t;
   print_section_variable    : env -> Evd.evar_map -> variable -> Pp.t;
-  print_syntactic_def       : env -> KerName.t -> Pp.t;
+  print_abbreviation        : env -> abbreviation -> Pp.t;
   print_module              : ModPath.t -> Pp.t;
   print_modtype             : ModPath.t -> Pp.t;
   print_named_decl          : env -> Evd.evar_map -> Constr.named_declaration -> Pp.t;
-  print_library_entry       : env -> Evd.evar_map -> bool -> (object_name * Lib.node) -> Pp.t option;
+  print_library_leaf       : env -> Evd.evar_map -> bool -> ModPath.t -> Libobject.t -> Pp.t option;
   print_context             : env -> Evd.evar_map -> bool -> int option -> Lib.library_segment -> Pp.t;
   print_typed_value_in_env  : Environ.env -> Evd.evar_map -> EConstr.constr * EConstr.types -> Pp.t;
   print_eval                : Reductionops.reduction_function -> env -> Evd.evar_map -> Constrexpr.constr_expr -> EConstr.unsafe_judgment -> Pp.t;
@@ -198,7 +198,11 @@ let print_opacity ref =
 (*******************)
 
 let print_if_is_coercion ref =
-  if Coercionops.coercion_exists ref then [pr_global ref ++ str " is a coercion"] else []
+  if Coercionops.coercion_exists ref then
+    let i = Coercionops.coercion_info ref in
+    let r = if i.Coercionops.coe_reversible then " reversible" else "" in
+    [pr_global ref ++ str " is a" ++ str r ++ str " coercion"]
+  else []
 
 (*******************)
 (* *)
@@ -389,7 +393,7 @@ type locatable = Locatable : 'a locatable_info -> locatable
 type logical_name =
   | Term of GlobRef.t
   | Dir of Nametab.GlobDirRef.t
-  | Syntactic of KerName.t
+  | Abbreviation of abbreviation
   | Module of ModPath.t
   | ModuleType of ModPath.t
   | Other : 'a * 'a locatable_info -> logical_name
@@ -406,7 +410,7 @@ exception ObjFound of logical_name
 let locate_any_name qid =
   try Term (Nametab.locate qid)
   with Not_found ->
-  try Syntactic (Nametab.locate_syndef qid)
+  try Abbreviation (Nametab.locate_abbreviation qid)
   with Not_found ->
   try Dir (Nametab.locate_dir qid)
   with Not_found ->
@@ -429,8 +433,8 @@ let pr_located_qualid = function
         | ConstructRef _ -> "Constructor"
         | VarRef _ -> "Variable" in
       str ref_str ++ spc () ++ pr_path (Nametab.path_of_global ref)
-  | Syntactic kn ->
-      str "Notation" ++ spc () ++ pr_path (Nametab.path_of_syndef kn)
+  | Abbreviation kn ->
+      str "Notation" ++ spc () ++ pr_path (Nametab.path_of_abbreviation kn)
   | Dir dir ->
       let s,dir =
         let open Nametab in
@@ -477,8 +481,8 @@ let locate_term qid =
   let expand = function
     | TrueGlobal ref ->
         Term ref, Nametab.shortest_qualid_of_global Id.Set.empty ref
-    | SynDef kn ->
-        Syntactic kn, Nametab.shortest_qualid_of_syndef Id.Set.empty kn
+    | Abbrev kn ->
+        Abbreviation kn, Nametab.shortest_qualid_of_abbreviation Id.Set.empty kn
   in
   List.map expand (Nametab.locate_extended_all qid)
 
@@ -641,7 +645,7 @@ let print_constant with_values sep sp udecl =
     | Some (c, priv, ctx) ->
       let priv = match priv with
       | Opaqueproof.PrivateMonomorphic () -> None
-      | Opaqueproof.PrivatePolymorphic (_, ctx) -> Some ctx
+      | Opaqueproof.PrivatePolymorphic ctx -> Some ctx
       in
         print_basename sp ++ print_instance sigma cb ++ str sep ++ cut () ++
         (if with_values then print_typed_body env sigma (Some c,typ) else pr_ltype typ)++
@@ -651,9 +655,9 @@ let gallina_print_constant_with_infos sp udecl =
   print_constant true " = " sp udecl ++
   with_line_skip (print_name_infos (GlobRef.ConstRef sp))
 
-let gallina_print_syntactic_def env kn =
-  let qid = Nametab.shortest_qualid_of_syndef Id.Set.empty kn
-  and (vars,a) = Syntax_def.search_syntactic_definition kn in
+let gallina_print_abbreviation env kn =
+  let qid = Nametab.shortest_qualid_of_abbreviation Id.Set.empty kn
+  and (vars,a) = Abbreviation.search_abbreviation kn in
   let c = Notation_ops.glob_constr_of_notation_constr a in
   hov 2
     (hov 4
@@ -662,7 +666,7 @@ let gallina_print_syntactic_def env kn =
         spc () ++ str ":=") ++
      spc () ++
      Constrextern.without_specific_symbols
-       [Notation.SynDefRule kn] (pr_glob_constr_env env (Evd.from_env env)) c)
+       [Notation_term.AbbrevRule kn] (pr_glob_constr_env env (Evd.from_env env)) c)
 
 module DynHandle = Libobject.Dyn.Map(struct type 'a t = 'a -> Pp.t option end)
 
@@ -674,55 +678,74 @@ let handle h (Libobject.Dyn.Dyn (tag, o)) = match DynHandle.find tag h with
    no reason that an object in the stack corresponds to a user-facing
    declaration. It may have been so at the time this was written, but this
    needs to be done in a more principled way. *)
-let gallina_print_leaf_entry env sigma with_values ((sp, kn),lobj) =
+let gallina_print_library_leaf env sigma with_values mp lobj =
   let sep = if with_values then " = " else " : " in
   match lobj with
   | AtomicObject o ->
     let handler =
-      DynHandle.add Declare.Internal.objVariable begin fun _ ->
+      DynHandle.add Declare.Internal.objVariable begin fun id ->
           (* Outside sections, VARIABLES still exist but only with universes
              constraints *)
-          (try Some(print_named_decl env sigma (basename sp)) with Not_found -> None)
+          (try Some(print_named_decl env sigma id) with Not_found -> None)
       end @@
-      DynHandle.add Declare.Internal.Constant.tag begin fun _ ->
-          Some (print_constant with_values sep (Constant.make1 kn) None)
+      DynHandle.add Declare.Internal.Constant.tag begin fun (id,_) ->
+        let kn = Constant.make2 mp (Label.of_id id) in
+        Some (print_constant with_values sep kn None)
       end @@
-      DynHandle.add DeclareInd.Internal.objInductive begin fun _ ->
-          Some (gallina_print_inductive (MutInd.make1 kn) None)
+      DynHandle.add DeclareInd.Internal.objInductive begin fun (id,_) ->
+        let kn = MutInd.make2 mp (Label.of_id id) in
+        Some (gallina_print_inductive kn None)
       end @@
       DynHandle.empty
     in
     handle handler o
-  | ModuleObject _ ->
-    let (mp,l) = KerName.repr kn in
-    Some (print_module ~with_body:with_values (MPdot (mp,l)))
-  | ModuleTypeObject _ ->
-    let (mp,l) = KerName.repr kn in
-          Some (print_modtype (MPdot (mp,l)))
-  | _ -> None
+  | ModuleObject (id,_) ->
+    Some (print_module ~with_body:with_values (MPdot (mp,Label.of_id id)))
+  | ModuleTypeObject (id,_) ->
+    Some (print_modtype (MPdot (mp, Label.of_id id)))
+  | IncludeObject _ | KeepObject _ | ExportObject _ -> None
 
-let gallina_print_library_entry env sigma with_values ent =
-  let pr_name (sp,_) = Id.print (basename sp) in
-  match ent with
-    | (oname,Lib.Leaf lobj) ->
-      gallina_print_leaf_entry env sigma with_values (oname,lobj)
-    | (oname,Lib.OpenedSection (dir,_)) ->
-      Some (str " >>>>>>> Section " ++ pr_name oname)
-    | (_,Lib.CompilingLibrary { Nametab.obj_dir; _ }) ->
-      Some (str " >>>>>>> Library " ++ DirPath.print obj_dir)
-    | (oname,Lib.OpenedModule _) ->
-      Some (str " >>>>>>> Module " ++ pr_name oname)
+let decr = Option.map ((+) (-1))
+
+let is_done = Option.equal Int.equal (Some 0)
+
+let print_leaves env sigma with_values mp =
+  let rec prec n = function
+    | [] -> n, mt()
+    | o :: rest ->
+      if is_done n then n, mt()
+      else begin match gallina_print_library_leaf env sigma with_values mp o with
+        | Some pp ->
+          let n, prest = prec (decr n) rest in
+          n, prest ++ pp
+        | None -> prec n rest
+      end
+  in
+  prec
 
 let gallina_print_context env sigma with_values =
   let rec prec n = function
-    | h::rest when Option.is_empty n || Option.get n > 0 ->
-      (match gallina_print_library_entry env sigma with_values h with
-       | None -> prec n rest
-       | Some pp -> prec (Option.map ((+) (-1)) n) rest ++ pp ++ fnl ())
-    | _ -> mt ()
+    | [] -> mt()
+    | (node, leaves) :: rest ->
+      if is_done n then mt()
+      else
+        let mp = (Lib.node_prefix node).Nametab.obj_mp in
+        let n, pleaves = print_leaves env sigma with_values mp n leaves in
+        if is_done n then pleaves
+        else prec n rest ++ pleaves
   in
   prec
 
+let pr_prefix_name prefix = Id.print (snd (split_dirpath prefix.Nametab.obj_dir))
+
+let print_library_node = function
+  | Lib.OpenedSection (prefix, _) ->
+    str " >>>>>>> Section " ++ pr_prefix_name prefix
+  | Lib.OpenedModule (_,_,prefix,_) ->
+    str " >>>>>>> Module " ++ pr_prefix_name prefix
+  | Lib.CompilingLibrary { Nametab.obj_dir; _ } ->
+    str " >>>>>>> Library " ++ DirPath.print obj_dir
+
 let gallina_print_eval red_fun env sigma _ {uj_val=trm;uj_type=typ} =
   let ntrm = red_fun env sigma trm in
   (str "     = " ++ gallina_print_typed_value_in_env env sigma (ntrm,typ))
@@ -734,11 +757,11 @@ let default_object_pr = {
   print_inductive           = gallina_print_inductive;
   print_constant_with_infos = gallina_print_constant_with_infos;
   print_section_variable    = gallina_print_section_variable;
-  print_syntactic_def       = gallina_print_syntactic_def;
+  print_abbreviation        = gallina_print_abbreviation;
   print_module              = gallina_print_module;
   print_modtype             = gallina_print_modtype;
   print_named_decl          = gallina_print_named_decl;
-  print_library_entry       = gallina_print_library_entry;
+  print_library_leaf        = gallina_print_library_leaf;
   print_context             = gallina_print_context;
   print_typed_value_in_env  = gallina_print_typed_value_in_env;
   print_eval                = gallina_print_eval;
@@ -750,11 +773,11 @@ let set_object_pr = (:=) object_pr
 let print_inductive x = !object_pr.print_inductive x
 let print_constant_with_infos c = !object_pr.print_constant_with_infos c
 let print_section_variable c = !object_pr.print_section_variable c
-let print_syntactic_def x = !object_pr.print_syntactic_def x
+let print_abbreviation x = !object_pr.print_abbreviation x
 let print_module x  = !object_pr.print_module  x
 let print_modtype x = !object_pr.print_modtype x
 let print_named_decl x = !object_pr.print_named_decl x
-let print_library_entry x = !object_pr.print_library_entry x
+let print_library_leaf x = !object_pr.print_library_leaf x
 let print_context x = !object_pr.print_context x
 let print_typed_value_in_env x = !object_pr.print_typed_value_in_env x
 let print_eval x = !object_pr.print_eval x
@@ -788,53 +811,61 @@ let handleF h (Libobject.Dyn.Dyn (tag, o)) = match DynHandleF.find tag h with
 | exception Not_found -> mt ()
 
 (* TODO: see the comment for {!gallina_print_leaf_entry} *)
+let print_full_pure_atomic env sigma mp lobj =
+  let handler =
+    DynHandleF.add Declare.Internal.Constant.tag begin fun (id,_) ->
+      let kn = KerName.make mp (Label.of_id id) in
+      let con = Global.constant_of_delta_kn kn in
+      let cb = Global.lookup_constant con in
+      let typ = cb.const_type in
+      hov 0 (
+        match cb.const_body with
+        | Undef _ ->
+          str "Parameter " ++
+          print_basename con ++ str " : " ++ cut () ++ pr_ltype_env env sigma typ
+        | OpaqueDef lc ->
+          str "Theorem " ++ print_basename con ++ cut () ++
+          str " : " ++ pr_ltype_env env sigma typ ++ str "." ++ fnl () ++
+          str "Proof " ++ pr_lconstr_env env sigma
+            (fst (Global.force_proof Library.indirect_accessor lc))
+        | Def c ->
+          str "Definition " ++ print_basename con ++ cut () ++
+          str "  : " ++ pr_ltype_env env sigma typ ++ cut () ++ str " := " ++
+          pr_lconstr_env env sigma c
+        | Primitive _ ->
+          str "Primitive " ++
+          print_basename con ++ str " : " ++ cut () ++ pr_ltype_env env sigma typ)
+      ++ str "." ++ fnl () ++ fnl ()
+    end @@
+    DynHandleF.add DeclareInd.Internal.objInductive begin fun (id,_) ->
+      let kn = KerName.make mp (Label.of_id id) in
+      let mind = Global.mind_of_delta_kn kn in
+      let mib = Global.lookup_mind mind in
+      pr_mutual_inductive_body (Global.env()) mind mib None ++
+      str "." ++ fnl () ++ fnl ()
+    end @@
+    DynHandleF.empty
+  in
+  handleF handler lobj
+
+let print_full_pure_leaf env sigma mp = function
+  | AtomicObject lobj -> print_full_pure_atomic env sigma mp lobj
+  | ModuleObject (id, _) ->
+    (* TODO: make it reparsable *)
+    print_module (MPdot (mp,Label.of_id id)) ++ str "." ++ fnl () ++ fnl ()
+  | ModuleTypeObject (id, _) ->
+    (* TODO: make it reparsable *)
+    print_modtype (MPdot (mp,Label.of_id id)) ++ str "." ++ fnl () ++ fnl ()
+  | _ -> mt()
+
 let print_full_pure_context env sigma =
   let rec prec = function
-  | ((_,kn),Lib.Leaf AtomicObject lobj)::rest ->
-    let handler =
-      DynHandleF.add Declare.Internal.Constant.tag begin fun _ ->
-          let con = Global.constant_of_delta_kn kn in
-          let cb = Global.lookup_constant con in
-          let typ = cb.const_type in
-          hov 0 (
-            match cb.const_body with
-              | Undef _ ->
-                str "Parameter " ++
-                print_basename con ++ str " : " ++ cut () ++ pr_ltype_env env sigma typ
-              | OpaqueDef lc ->
-                str "Theorem " ++ print_basename con ++ cut () ++
-                str " : " ++ pr_ltype_env env sigma typ ++ str "." ++ fnl () ++
-                str "Proof " ++ pr_lconstr_env env sigma
-                  (fst (Global.force_proof Library.indirect_accessor lc))
-              | Def c ->
-                str "Definition " ++ print_basename con ++ cut () ++
-                str "  : " ++ pr_ltype_env env sigma typ ++ cut () ++ str " := " ++
-                pr_lconstr_env env sigma c
-              | Primitive _ ->
-                 str "Primitive " ++
-                   print_basename con ++ str " : " ++ cut () ++ pr_ltype_env env sigma typ)
-          ++ str "." ++ fnl () ++ fnl ()
-      end @@
-      DynHandleF.add DeclareInd.Internal.objInductive begin fun _ ->
-          let mind = Global.mind_of_delta_kn kn in
-          let mib = Global.lookup_mind mind in
-          pr_mutual_inductive_body (Global.env()) mind mib None ++
-            str "." ++ fnl () ++ fnl ()
-      end @@
-      DynHandleF.empty
-    in
-    let pp = handleF handler lobj in
+    | (node,leaves)::rest ->
+      let mp = (Lib.node_prefix node).Nametab.obj_mp in
+      let pp = Pp.prlist (print_full_pure_leaf env sigma mp) leaves in
       prec rest ++ pp
-  | ((_,kn),Lib.Leaf ModuleObject _)::rest ->
-          (* TODO: make it reparsable *)
-    let (mp,l) = KerName.repr kn in
-          prec rest ++ print_module (MPdot (mp,l)) ++ str "." ++ fnl () ++ fnl ()
-  | ((_,kn),Lib.Leaf ModuleTypeObject _)::rest ->
-          (* TODO: make it reparsable *)
-    let (mp,l) = KerName.repr kn in
-          prec rest ++ print_modtype (MPdot (mp,l)) ++ str "." ++ fnl () ++ fnl ()
-  | _::rest -> prec rest
-  | _ -> mt () in
+  | [] -> mt ()
+  in
   prec (Lib.contents ())
 
 (* For printing an inductive definition with
@@ -849,7 +880,7 @@ let read_sec_context qid =
     with Not_found ->
       user_err ?loc:qid.loc (str "Unknown section.") in
   let rec get_cxt in_cxt = function
-    | (_,Lib.OpenedSection ({Nametab.obj_dir;_},_) as hd)::rest ->
+    | (Lib.OpenedSection ({Nametab.obj_dir;_},_), _ as hd)::rest ->
         if DirPath.equal dir obj_dir then (hd::in_cxt) else get_cxt (hd::in_cxt) rest
     | [] -> []
     | hd::rest -> get_cxt (hd::in_cxt) rest
@@ -867,7 +898,7 @@ let maybe_error_reject_univ_decl na udecl =
   let open GlobRef in
   match na, udecl with
   | _, None | Term (ConstRef _ | IndRef _ | ConstructRef _), Some _ -> ()
-  | (Term (VarRef _) | Syntactic _ | Dir _ | Module _ | ModuleType _ | Other _ | Undefined _), Some udecl ->
+  | (Term (VarRef _) | Abbreviation _ | Dir _ | Module _ | ModuleType _ | Other _ | Undefined _), Some udecl ->
     (* TODO Print na somehow *)
     user_err (str "This object does not support universe names.")
 
@@ -879,7 +910,7 @@ let print_any_name env sigma na udecl =
   | Term (IndRef (sp,_)) -> print_inductive sp udecl
   | Term (ConstructRef ((sp,_),_)) -> print_inductive sp udecl
   | Term (VarRef sp) -> print_section_variable env sigma sp
-  | Syntactic kn -> print_syntactic_def env kn
+  | Abbreviation kn -> print_abbreviation env kn
   | Module mp -> print_module mp
   | Dir _ -> mt ()
   | ModuleType mp -> print_modtype mp
@@ -902,6 +933,47 @@ let print_name env sigma na udecl =
   | {loc; v=Constrexpr.AN ref} ->
     print_any_name env sigma (locate_any_name ref) udecl
 
+let print_notation_grammar env sigma ntn =
+  let ng = List.hd (Notgram_ops.grammar_of_notation ntn) in
+  let assoc = ng.Notation_gram.notgram_assoc in
+  let prdf () = Pp.str "no associativity" in
+  Pp.(pr_opt_no_spc_default prdf Gramlib.Gramext.pr_assoc assoc)
+
+exception PrintNotationNotFound of Constrexpr.notation_entry * string
+
+let () = CErrors.register_handler @@ function
+  | PrintNotationNotFound (entry, ntn_str) ->
+      let entry_string = match entry with
+      | Constrexpr.InConstrEntry -> "."
+      | Constrexpr.InCustomEntry e -> " in " ^ e ^ " entry."
+      in
+      Some Pp.(str "\"" ++ str ntn_str ++ str "\"" ++ spc ()
+        ++ str "cannot be interpreted as a known notation" ++ str entry_string ++ spc ()
+        ++ strbrk "Make sure that symbols are surrounded by spaces and that holes are explicitly denoted by \"_\".")
+  | _ -> None
+
+let error_print_notation_not_found e s =
+  raise @@ PrintNotationNotFound (e, s)
+
+let print_notation env sigma entry raw_ntn =
+  (* make sure entry exists *)
+  let () =
+    match entry with
+    | Constrexpr.InConstrEntry -> ()
+    | Constrexpr.InCustomEntry e -> Metasyntax.check_custom_entry e
+  in
+  (* convert notation string to key. eg. "x + y" to "_ + _" *)
+  let interp_ntn = Notation.interpret_notation_string raw_ntn in
+  let ntn = (entry, interp_ntn) in
+  try
+    let lvl = Notation.level_of_notation ntn in
+    let args = Notgram_ops.subentries_of_notation ntn in
+    let pplvl = Metasyntax.pr_level ntn lvl args in
+    Pp.(str "Notation \"" ++ str interp_ntn ++ str "\"" ++ spc ()
+      ++ pplvl ++ pr_comma () ++ print_notation_grammar env sigma ntn
+      ++ str ".")
+  with Not_found -> error_print_notation_not_found entry raw_ntn
+
 let print_opaque_name env sigma qid =
   let open GlobRef in
   match Nametab.global qid with
@@ -935,12 +1007,12 @@ let print_about_any ?loc env sigma k udecl =
         print_opacity ref @
   print_bidi_hints ref @
   [hov 0 (str "Expands to: " ++ pr_located_qualid k)])
-  | Syntactic kn ->
-    let () = match Syntax_def.search_syntactic_definition kn with
+  | Abbreviation kn ->
+    let () = match Abbreviation.search_abbreviation kn with
     | [],Notation_term.NRef (ref,_) -> Dumpglob.add_glob ?loc ref
     | _ -> () in
       v 0 (
-      print_syntactic_def env kn ++ fnl () ++
+      print_abbreviation env kn ++ fnl () ++
       hov 0 (str "Expands to: " ++ pr_located_qualid k))
   | Dir _ | Module _ | ModuleType _ | Undefined _ ->
       hov 0 (pr_located_qualid k)
@@ -970,7 +1042,8 @@ let print_path ((i,j),p) =
   hov 2 (
     str"[" ++ hov 0 (prlist_with_sep pr_semicolon print_coercion_value p) ++
     str"] : ") ++
-  pr_class i ++ str" >-> " ++ pr_class j
+  pr_class i ++ str" >-> " ++ pr_class j ++
+  str (if path_is_reversible p then " (reversible)" else "")
 
 let _ = Coercionops.install_path_printer print_path
 
