@@ -60,6 +60,7 @@ type printable =
   | PrintAssumptions of bool * bool * qualid or_by_notation
   | PrintStrategy of qualid or_by_notation option
   | PrintRegistered
+  | PrintNotation of Constrexpr.notation_entry * string
 
 type glob_search_where = InHyp | InConcl | Anywhere
 
@@ -107,13 +108,15 @@ type verbose_flag   = bool (* true = Verbose;       false = Silent         *)
 type coercion_flag  = bool (* true = AddCoercion    false = NoCoercion     *)
 type instance_flag  = BackInstance | NoInstance
 
-type export_flag    = bool (* true = Export;        false = Import         *)
+type export_flag = Lib.export_flag = Export | Import
 
 type import_categories = {
   negative : bool;
   import_cats : string CAst.t list;
 }
 
+type export_with_cats = export_flag * import_categories option
+
 type infix_flag     = bool (* true = Infix;         false = Notation       *)
 
 type one_import_filter_name = qualid * bool (* import inductive components *)
@@ -182,6 +185,7 @@ type 'a with_coercion = coercion_flag * 'a
 (* Attributes of a record field declaration *)
 type record_field_attr = {
   rf_subclass: instance_flag; (* the projection is an implicit coercion or an instance *)
+  rf_reverse: bool option;
   rf_priority: int option; (* priority of the instance, if relevant *)
   rf_notation: decl_notation list;
   rf_canonical: bool; (* use this projection in the search for canonical instances *)
@@ -189,7 +193,7 @@ type record_field_attr = {
 type constructor_expr = (lident * constr_expr) with_coercion
 type constructor_list_or_record_decl_expr =
   | Constructors of constructor_expr list
-  | RecordDecl of lident option * (local_decl_expr * record_field_attr) list
+  | RecordDecl of lident option * (local_decl_expr * record_field_attr) list * lident option
 type inductive_params_expr = local_binder_expr list * local_binder_expr list option
 (** If the option is nonempty the "|" marker was used *)
 
@@ -214,10 +218,22 @@ type proof_end =
   (*                         name in `Save ident` when closing goal *)
   | Proved of opacity_flag * lident option
 
-type scheme =
-  | InductionScheme of bool * qualid or_by_notation * Sorts.family
-  | CaseScheme of bool * qualid or_by_notation * Sorts.family
-  | EqualityScheme of qualid or_by_notation
+type scheme_type =
+  | SchemeInduction
+  | SchemeMinimality
+  | SchemeElimination
+  | SchemeCase
+
+type equality_scheme_type =
+  | SchemeBooleanEquality
+  | SchemeEquality
+
+  (* The data of a Scheme decleration *)
+type scheme = {
+  sch_type : scheme_type ;
+  sch_qualid : Libnames.qualid Constrexpr.or_by_notation ;
+  sch_sort : Sorts.family ;
+}
 
 type section_subset_expr =
   | SsEmpty
@@ -264,7 +280,7 @@ type register_kind =
 (** {6 Types concerning the module layer} *)
 
 type module_ast_inl = module_ast * Declaremods.inline
-type module_binder = bool option * lident list * module_ast_inl
+type module_binder = export_with_cats option * lident list * module_ast_inl
 
 (** {6 The type of vernacular expressions} *)
 
@@ -342,6 +358,7 @@ type nonrec vernac_expr =
   | VernacFixpoint of discharge * fixpoint_expr list
   | VernacCoFixpoint of discharge * cofixpoint_expr list
   | VernacScheme of (lident option * scheme) list
+  | VernacSchemeEquality of equality_scheme_type * Libnames.qualid Constrexpr.or_by_notation
   | VernacCombinedScheme of lident * lident list
   | VernacUniverse of lident list
   | VernacConstraint of univ_constraint_expr list
@@ -349,12 +366,13 @@ type nonrec vernac_expr =
   (* Gallina extensions *)
   | VernacBeginSection of lident
   | VernacEndSegment of lident
+  | VernacExtraDependency of qualid * string * Id.t option
   | VernacRequire of
-      qualid option * export_flag option * qualid list
-  | VernacImport of export_flag * import_categories option * (qualid * import_filter_expr) list
+      qualid option * export_with_cats option * (qualid * import_filter_expr) list
+  | VernacImport of export_with_cats * (qualid * import_filter_expr) list
   | VernacCanonical of qualid or_by_notation
   | VernacCoercion of qualid or_by_notation *
-      class_rawexpr * class_rawexpr
+      (class_rawexpr * class_rawexpr) option
   | VernacIdentityCoercion of lident * class_rawexpr * class_rawexpr
   | VernacNameSectionHypSet of lident * section_subset_expr
 
@@ -380,9 +398,9 @@ type nonrec vernac_expr =
   | VernacExistingClass of qualid (* inductive or definition name *)
 
   (* Modules and Module Types *)
-  | VernacDeclareModule of bool option * lident *
+  | VernacDeclareModule of export_with_cats option * lident *
       module_binder list * module_ast_inl
-  | VernacDefineModule of bool option * lident * module_binder list *
+  | VernacDefineModule of export_with_cats option * lident * module_binder list *
       module_ast_inl Declaremods.module_signature * module_ast_inl list
   | VernacDeclareModuleType of lident *
       module_binder list * module_ast_inl list * module_ast_inl list
@@ -436,7 +454,7 @@ type nonrec vernac_expr =
   | VernacComments of comment list
 
   (* Proof management *)
-  | VernacAbort of lident option
+  | VernacAbort
   | VernacAbortAll
   | VernacRestart
   | VernacUndo of int
